Require Import ssreflect fintype ssrbig matrix ssrnat. 
(*ssrbool funs eqtype ssrnat seq fintype.
Require Import tuple paths connect div .
Require Import groups group_perm zp ssralg ssrbig matrix.
*)
Require Import structRp min_max.
Require Import Reals.
Require Import Fourier.
Require Import Ranalysis.
Require Import Classical.
Require Import ClassicalChoice.
Require Import vectors seqRp matrixRp matrixNorm.
Require Import Max.

Open Local Scope R_scope.

Definition limit_gen (X X':Metric_Space) (f:Base X -> Base X')
   (x0:Base X) (l:Base X') :=
  forall eps:R,
    eps > R0 ->
    exists alp : R,
      alp > R0 /\
      (forall x:Base X, R0<dist X x x0 < alp -> dist X' (f x) l < eps).


Definition cont_gen (X X':Metric_Space) (f:Base X -> Base X')
   (x0:Base X)  :=
  forall eps:R,
    eps > R0 ->
    exists alp : R,
      alp > R0 /\
      (forall x:Base X, dist X x x0 < alp -> dist X' (f x) (f x0) < eps).

Lemma echiv_lim_f:
forall X X' f g x l,
(forall t, f t = g t) ->
(limit_gen X X' f x l <-> limit_gen X X' g x l).
Proof.
move => X X' f g x l H.
split; move => H1 eps He; elim (H1 eps He) =>a [Ha Hr] ;
exists a; split.
done.
by move => y Hy; rewrite -(H y); apply Hr.
done.
by move => y Hy; rewrite (H y) ; apply Hr.
Qed.

Lemma echiv_lim_f1:
forall X X' f g x l,
(forall t, f t = g t) ->
(limit_gen X X' f x l -> limit_gen X X' g x l).
Proof.
move => X X' f g x l H  H2.
by rewrite -(echiv_lim_f X X' f g x l H ).
Qed.

Lemma cond_eq_el :
  forall (X:Metric_Space) (x y:Base X), (forall eps:R, R0 < eps -> dist X x  y < eps) -> x = y.
Proof.
move =>X x y H.
have Hra: ~ R0 < dist X x y.
rewrite /not => Hdist.
have Hco: dist X x  y < (dist X x y)*/2.
by apply H; fourier.
have Hco2: ~ dist X x y < dist X x y */2.
by fourier.
contradiction.
rewrite -(Rlimit.dist_refl X).
by elim (Rlimit.dist_pos X x y).
Qed.

(*
Lemma uniq_lim:
forall X X' f x l k,
limit_gen X X' f x l -> 
limit_gen X X' f x k ->
l=k.
Proof.
move => X X' f x l k Hl Hk.
rewrite /limit_gen in Hl Hk.
have H: forall eps, eps>0-> dist X' l k <eps.
move => eps Heps.
elim (Hl (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp1 [Ha1 Hdis1].
elim (Hk (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp2 [Ha2 Hdis2].


have H: ( l<>k -> False).
move => Hdif.
have H1: (dist X' l k) > 0.
have H2 := Rlimit.dist_pos X' l k.
elim H2; [done|].
rewrite Rlimit.dist_refl; move => t; elim (Hdif t).
elim (Hl (dist X' l k) H1 ) => al [Hal Hd1].
elim (Hk (dist X' l k) H1 ) => ak [Hak Hd2].


intuition.
fold not  in H.


have H: forall eps, eps>0-> dist X' l k <eps.
move => eps Heps.
elim (Hl (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp1 [Ha1 Hdis1].
elim (Hk (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp2 [Ha2 Hdis2].

have He2:  
*)



Definition limit_R f x0 l:=
limit_gen R_met R_met f x0 l.

Definition cont_R f  x0 :=
cont_gen R_met R_met f  x0 .


Section funp.
Variable p: pos_nat.

Lemma Hp: (O<p)%nat.
Proof.
have Hp1:= pos_natP p.
move/leP:Hp1=>Hp1.
omega.
Qed.




(*functions (Rvec p)->(Rvec p) ; we assume total functions*)

Definition add_f (f1 f2:(Rvec p)->(Rvec p)) v:= add_v (f1 v) (f2 v).
Definition dif_f (f1 f2:(Rvec p)->(Rvec p)) v:= dif_v (f1 v) (f2 v).
Definition mult_sf (x:R) (f : (Rvec p) -> (Rvec p)) v:= mult_sv x (f v).

Definition zero_f (v:(Rvec p)):= vect0.

Lemma annul_f :
forall f v, mult_sf R0 f v = (vect0 p). 
Proof.
move =>*;
by rewrite -ext_eq; rewrite  /mult_sf /mult_sv /vect0 => i; 
rewrite !vxK ; ring_simplify.
Qed.


(*function limit; we do not define the limit with respect to a set as we have total functions*)
(*we need concepts for diferent types of functions R-> Rp, Rp -> R,
Rp -> Rp*)

Lemma dist_pos1 :
forall x y : (Rvec p), dist_Rp x y >= R0.
Proof.
move => x y.
apply Rle_ge; apply (dist_pos ).
Qed.

Definition Rp_met: Metric_Space:=
Build_Metric_Space 
(Rvec p) (@dist_Rp p) dist_pos1 (@dist_sym p) (@dist_refl p Hp) (@dist_tri p Hp).


Definition limit_Rp_R f  x0 l :=
limit_gen Rp_met R_met f  x0 l.

Definition limit_R_Rp f  x0 l :=
limit_gen R_met Rp_met f  x0 l.

Definition limit_Rp f x0 l:=
limit_gen Rp_met Rp_met f x0 l.

Definition cont_Rp_R f  x0 :=
cont_gen Rp_met R_met f  x0 .

Definition cont_R_Rp f  x0  :=
cont_gen R_met Rp_met f  x0 .

Definition cont_Rp f x0 :=
cont_gen Rp_met Rp_met f x0.


(*

Lemma echiv_limRp :
forall f v l,
limit_Rp f v l <-> limitRp_Rp f v l.
Proof.
move => f v l.
rewrite /limit_Rp /limitRp_Rp /limit_gen.
split; move =>H eps He; 
elim (H eps He) => alp Ha;
 elim Ha => Ha1 Hr; exists alp; split; first by done.
move => x  H1. by apply Hr.
done.
move => x H1; by apply Hr.
Qed.

Lemma echiv_lim_Rp_R:
forall f v l,
limit_Rp_R f v l <-> limit_Rp_1 f v l.
Proof.
move => f v l.
rewrite /limit_Rp_R /limit_Rp_1 /limit_gen.
split; move =>H eps He; 
elim (H eps He) => alp Ha;
 elim Ha => Ha1 Hr; exists alp; split; first by done.
move => x H1.
by rewrite /dist //= /R_dist; apply Hr.
done.
by move => x H1;  rewrite /dist //= /R_dist in Hr ;apply Hr.
Qed.



Definition limit_Rp (f:(Rvec p) ->(Rvec p)) (v0:(Rvec p)) (l:(Rvec p)):=
forall eps:R, eps>R0-> 
       exists alp:R, alp>R0/\(forall v:(Rvec p),R0< dist_Rp v v0<alp -> dist_Rp (f v) l<eps).
*)

Lemma limit_vect0:
forall (f:(Rvec p)->(Rvec p))(v:(Rvec p)),
(forall v0, f v0 = vect0 p) -> limit_Rp f v (vect0 p).
Proof.
move=> f v Heq.
rewrite /limit_Rp => eps Heps.
exists eps.
split; [assumption|move => v0 Hdist].
rewrite (Heq v0).
have Heq0: dist_Rp (vect0 p) (vect0 p) = R0 by rewrite (dist_refl Hp).
by rewrite /dist //= Heq0.
Qed.

Lemma limit_Rp_plus:
forall f1 f2 l1 l2 v,
limit_Rp f1 v l1 ->
limit_Rp f2 v l2 ->
limit_Rp (add_f f1 f2) v (add_v l1 l2).
Proof.
move => f1 f2 l1 l2 v Hl1 Hl2.
rewrite / limit_Rp /add_f /dist_Rp =>eps Heps.
rewrite /limit_Rp in Hl1 Hl2.
elim (Hl1 (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp1 [Ha1 Hdis1].
elim (Hl2 (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp2 [Ha2 Hdis2].
exists (Rmin alp1 alp2).
split; [|move => v0 Hineq].
unfold Rmin; case (Rle_dec alp1 alp2) => H;fourier.
have Heq: dif_v (add_v (f1 v0 ) (f2 v0)) (add_v l1 l2) = add_v (dif_v (f1 v0) l1) (dif_v (f2 v0) l2).
 rewrite - ext_eq; rewrite /dif_v /add_v => i;rewrite !vxK. 
rewrite //=;field.
rewrite /dist //= /dist_Rp Heq.
apply Rle_lt_trans with (norm (dif_v (f1 v0) l1) + (norm (dif_v (f2 v0) l2))).
apply (norm_tri Hp).
fold (dist_Rp (f1 v0) l1) (dist_Rp (f2 v0) l2) .
rewrite /dist //= in Hineq.
elim Hineq  => H0 Hmin.
have Hin1: R0 < dist_Rp v0 v < alp1. 
split; [assumption|].
by destruct (Rmin_Rgt_l alp1 alp2 (dist_Rp v0 v)); fourier.
have Hin2: R0 < dist_Rp v0 v < alp2. 
split; [assumption|].
destruct (Rmin_Rgt_l alp1 alp2 (dist_Rp v0 v)); fourier.
replace eps with (eps*/2 + eps*/2)%R; [|field].
apply Rplus_lt_compat.
exact (Hdis1 v0 Hin1).
exact (Hdis2 v0 Hin2).
Qed.

Lemma limit_R_multc:
forall f k x l,
limit_R f x l ->
limit_R (fun t => k*f t) x (k*l).
Proof.
move => f k x l Hl eps He.
elim (Req_dec k R0) => Hk0.
exists eps; split; first by done.
by move => x0 Hx0; rewrite Hk0 !Rmult_0_l /dist //= /R_dist Rminus_0_r Rabs_R0.
have Hek: /Rabs k*eps > R0.
apply: Rmult_gt_0_compat ; [by apply: Rinv_0_lt_compat; apply Rabs_pos_lt|done].
elim (Hl (/Rabs k*eps) Hek) => del [Hd Hr].
exists del; split; first by done.
move => x0 Hx0.
rewrite /dist //= /R_dist -Rmult_minus_distr_l Rabs_mult.
apply (Rmult_lt_reg_l (/Rabs k)).
by apply: Rinv_0_lt_compat; apply Rabs_pos_lt.
ring_simplify.
rewrite Rinv_l; [| by apply Rgt_not_eq; apply: Rabs_pos_lt].
 rewrite Rmult_1_l.
by apply: Hr.
Qed.


Lemma limit_Rp_uniq:
forall f v l l', limit_Rp f v l -> limit_Rp f v l' -> l = l'.
Proof.
move => f v l l' Hll Hll'.
rewrite /limit_Rp in Hll Hll'.
rewrite -(dist_refl Hp).
have Hra: (~ (R0<dist_Rp l l')).
rewrite /not => Hdist.
elim (Hll ((dist_Rp l l')*/2) (eps2_Rgt_R0 (dist_Rp l l') Hdist)) => alp1 [Ha1 Hdis1].
elim (Hll' ((dist_Rp l l')*/2) (eps2_Rgt_R0 (dist_Rp l l') Hdist)) => alp2 [Ha2 Hdis2].
have Heq1: dist_Rp l l' = norm (dif_v l l') by done.
have Heq: forall u, dif_v l l'  = add_v (dif_v l (f u)) (dif_v (f u) l').
move => u; rewrite - ext_eq; rewrite  /dif_v /add_v =>i; rewrite !vxK;
rewrite //=; field.
have Hex: exists x, R0 < dist_Rp x v < Rmin alp1 alp2.
unfold dist_Rp.
have Hex2: R0< (Rmin alp1 alp2)*/2.
unfold Rmin; case Rle_dec; move=>*; by  apply :eps2_Rgt_R0.
 have Hex3: R0< (Rmin alp1 alp2)*/2< Rmin alp1 alp2. 
split; [done| apply Rlt_eps2_eps; fourier].
elim (ex_v_norm p Hp ((Rmin alp1 alp2)*/2) Hex2) => ev Heq2.
exists (add_v ev v).
have Heqd: dif_v (add_v ev v) v = ev.
 rewrite - ext_eq ; rewrite  /dif_v /add_v=>i; rewrite !vxK  //=;
 field.
by rewrite Heqd Heq2.
elim Hex => x Hine.
have Hco: norm (add_v (dif_v l (f x)) (dif_v (f x) l')) < dist_Rp l l'.
apply Rle_lt_trans with (norm (dif_v l (f x)) + norm (dif_v (f x) l')).
apply (norm_tri Hp).
replace (dist_Rp l l') with (dist_Rp l l' */2 +dist_Rp l l' */2)%R;[|field]. 
elim Hine => Hmin0 Hmin1.
have Hin1: R0 < dist_Rp x v < alp1. 
split; [assumption|].
destruct (Rmin_Rgt_l alp1 alp2 (dist_Rp x v)); fourier.
have Hin2: R0 < dist_Rp x v < alp2. 
split; [assumption|].
destruct (Rmin_Rgt_l alp1 alp2 (dist_Rp x v)); fourier.
apply Rplus_lt_compat.
fold (dist_Rp l (f x)).
rewrite {1}dist_sym.
exact (Hdis1 x Hin1). 
fold (dist_Rp (f x) l').
exact (Hdis2 x Hin2).
rewrite -(Heq x)  /dist_Rp in Hco.
have Hco2: (~( norm (dif_v l l')< norm (dif_v l l'))); fourier.
by elim (dist_pos  l l') => H.
Qed.

(*
(*continuity*)

Definition cont_Rp f v0:=
limit_Rp f v0 (f v0). 
*)
(*linear functions*)

Definition linear (f: Rvec p -> (Rvec p)) :Prop:=
forall x y, add_v (f x) (f y) = f (add_v x y)/\
forall x c, mult_sv c (f x) = f (mult_sv c x).


Lemma Heine1_Rp:
forall f Un x,
cont_Rp f x -> conv_Rp Un x -> conv_Rp (fun i:nat=>f (Un i)) (f x).
Proof.
move => f Un x H H0 eps0 He.
destruct (H eps0 He) as [delta [H1 H2]] .
destruct (H0 delta H1) as [N H3] .
exists N.
intros n Hn.
elim (dist_pos (Un n) x) => H4.
apply (H2 (Un n)).
exact (H3 n Hn).
have Heq: Un n = x.
by rewrite -(dist_refl Hp).
by rewrite Heq (dist0_eq p Hp).
Qed.

Lemma same_lim:
forall xn (x:Rvec p),
(conv_Rp xn x)->
(conv_Rp (fun i:nat => xn (S i)) x).
intros.
unfold conv_Rp in H |-*.
intros.
elim (H eps H0); intros N Hxn.
exists (N-(S O))%nat.
intros.
apply (Hxn (S n)).
omega.
Qed.
 
Lemma cv_dif_Rp:
forall xn yn (x y:Rvec p) ,
conv_Rp xn x ->
conv_Rp yn y ->
conv_Rp (fun n:nat => dif_v (xn n) (yn n)) (dif_v x y).
Proof.
move => xn yn x y Hcvx Hcvy eps He.
elim (Hcvx (eps*/2)(eps2_Rgt_R0 eps He)) => N1 Hcvx1.
elim (Hcvy (eps*/2)(eps2_Rgt_R0 eps He)) => N2 Hcvy1.
exists (max N1 N2) => n Hineq.
apply Rle_lt_trans with (dist_Rp (xn n) x + dist_Rp y (yn n)).
rewrite /dist_Rp.
replace (dif_v (dif_v (xn n) (yn n)) (dif_v x y)) with (add_v (dif_v (xn n) x) (dif_v y (yn n))).
apply (norm_tri Hp).
by rewrite - ext_eq; rewrite  /dif_v /add_v => i; rewrite !vxK  //=; ring_simplify.
have H1: (N1<=n)%nat. 
elim (max_dec N1 N2) => H.
by rewrite H in Hineq.
apply le_trans with (max N1 N2);[apply le_max_l|done] .
have H2: (N2<=n)%nat. 
elim (max_dec N1 N2) => H.
apply le_trans with (max N1 N2);[apply le_max_r|done] .
by rewrite H in Hineq.
replace eps with (eps*/2+eps*/2)%R; [|field].
apply Rplus_lt_compat.
apply (Hcvx1 n H1).
rewrite dist_sym.
apply (Hcvy1 n H2).
Qed.

Lemma lim_minus_0:
forall xn x,
(conv_Rp xn x)->
(conv_Rp (fun i:nat => dif_v (xn i) (xn (S i))) (vect0 p)).
Proof.
move => xn  x Hcv.
unfold conv_Rp => eps He .
replace (vect0 p) with (dif_v x x).
apply cv_dif_Rp.
exact Hcv.
by apply same_lim.
exact He.
rewrite - ext_eq; rewrite  /vect0 /dif_v => i; rewrite !vxK  //=
; by ring_simplify.
Qed.

(*some relations between dif concepts of limits
e.g. the limit in Rp is a limit on components etc. *)

Lemma echiv_lim_R_Rp:
forall f a l,
limit_R_Rp f a l <-> forall i, limit_R (fun v => (f v) i) a (l i).
Proof.
move => f a l; split => H; 
rewrite /limit_R /limit_R_Rp /limit_gen /dist //= /dist_Rp /R_dist in H |-*.
move => i eps He ; elim (H eps He) => alp [Ha Hr].
exists alp; split; first by done.
move =>x Hxa.
have Hr1 := (Hr x Hxa).
apply Rle_lt_trans with (norm (dif_v (f x) l)).
replace (f x i - l i) with ((dif_v (f x) l) i) ;[|by rewrite vxK].
apply: norm_max.
exact: Hr1.
move => eps He.
have Hl1: forall eps, eps > R0 -> forall i, exists alp, alp > R0 /\
            (forall x :  R,
             R0 < Rabs (x - a) < alp -> Rabs (f x i - l i)  < eps).
move=> e Hep i; exact (H i e Hep).
have Hep: eps/INR p > R0.
have Hp1: R0<INR p.
 apply lt_INR_0; exact Hp.
rewrite /Rdiv.
 apply Rmult_gt_0_compat; [done| ].
by apply: Rinv_0_lt_compat; apply lt_INR_0; exact Hp.

elim (choice (fun i alp => alp > R0 /\
          (forall x : R,
           R0 < Rabs (x - a) < alp -> Rabs (f x i - l i) < eps)) (Hl1 (eps) He))=> cf Hrel.
exists (min_del cf) .
split.
apply: (min_ge0 Hp).
move => j; elim (Hrel j) => Hinq Hrest; apply: Hinq.
move =>x Hdist.
elim (norm_rew Hp (dif_v (f x) l)) => k Hk;
rewrite Hk vxK.
elim (Hrel k) => H1 H2.
apply (H2 x).
split; elim Hdist => Hd1 Hd2; first by done.
apply Rlt_le_trans with (min_del cf); [done |apply min_min].
Qed.

Lemma uniq_lim_R:
forall f x l k,
limit_R  f x l -> 
limit_R f x k ->
l=k.
Proof.
move => f x l k Hl Hk.
rewrite /limit_R /limit_gen /dist //= in Hl Hk.
have H: forall eps, eps>R0-> R_dist l k <eps.
move => eps Heps.
elim (Hl (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp1 [Ha1 Hdis1].
elim (Hk (eps * / 2) (eps2_Rgt_R0 eps Heps))=> alp2 [Ha2 Hdis2].
replace eps with (eps */2 + eps */2) ; [|by field].
have Hmin: Rmin alp1 alp2 >R0 by apply Rmin_Rgt_r.
have H1: R0 < R_dist (x+ (Rmin alp1 alp2)*/2) x < alp1.
rewrite /R_dist /Rminus.
rewrite Rplus_comm -Rplus_assoc  Rplus_opp_l Rplus_0_l Rabs_right.
split; [by apply eps2_Rgt_R0|].
apply Rlt_le_trans with (Rmin alp1 alp2);[by apply Rlt_eps2_eps| by apply Rmin_l].
by rewrite /Rge; left; apply : eps2_Rgt_R0.
have H2: R0 < R_dist (x+ (Rmin alp1 alp2)*/2) x < alp2.
rewrite /R_dist /Rminus.
rewrite Rplus_comm -Rplus_assoc  Rplus_opp_l Rplus_0_l Rabs_right.
split; [by apply eps2_Rgt_R0|].
apply Rlt_le_trans with (Rmin alp1 alp2);[by apply Rlt_eps2_eps| by apply Rmin_r].
by rewrite /Rge; left; apply : eps2_Rgt_R0.
have Hd1:= Hdis1 (x + Rmin alp1 alp2 * / 2) H1.
have Hd2 := Hdis2 (x + Rmin alp1 alp2 * / 2) H2.
apply Rle_lt_trans with (R_dist (f (x + Rmin alp1 alp2 * / 2)) l + R_dist (f (x + Rmin alp1 alp2 * / 2)) k).
by rewrite (R_dist_sym (f (x + Rmin alp1 alp2 * / 2)) l ) ; apply R_dist_tri.
by apply Rplus_lt_compat.
by apply cond_eq_el.
Qed.

Notation "'MR_' p" := (matrix real_field p p)(at level 10).
Notation "\vec_ ( i ) E" :=
  (@vec_of_fun p (fun i  => E)) (at level 25).


Lemma cont_norm_matr:
forall (f: Rvec p->MR_(p)) (v:Rvec p),
(forall i, cont_Rp (fun v0:Rvec p  => \vec_(j) (f v0 i j)) v) ->
(forall eps,R0<eps-> exists del, R0<del /\
  forall v0, R0< dist_Rp v0 v < del -> norm_m (dif_m (f v0) (f v))<eps).
Proof.
move => f v Hcont eps Heps.
rewrite /cont_Rp /cont_gen in Hcont.
have Hc1: forall eps, eps >R0 -> forall i, exists alp, alp > R0 /\
            (forall v0 :Rvec p,
             R0 < dist_Rp v0 v < alp -> 
             dist_Rp (\vec_(j) (f v0 i j)) (\vec_(j) (f v i j)) < eps).
move=> e He i.
elim (Hcont i e He) => alp [Ha Hc].
exists alp.
split; first by done.
move => v0 [H1 H2]. by apply (Hc v0).
have Hep: eps/INR p >R0.
have Hp1: R0<INR p.
by apply lt_INR_0; exact Hp.
rewrite /Rdiv.
apply Rmult_gt_0_compat.
fourier.
have Haux : R0 < / INR p.
by apply Rinv_0_lt_compat.
auto with real.
elim (choice (fun i alp => alp > R0 /\
          (forall v0 : Rvec p,
           R0 < dist_Rp v0 v < alp -> 
          dist_Rp (\vec_(j) (f v0 i j)) (\vec_(j) (f v i j)) < eps/INR p)) (Hc1 (eps/INR p) Hep))=> cf Hrel.
exists (min_del cf) .
split.
apply (min_ge0 Hp); move => j; elim (Hrel j) => H1 _; done.
move =>v0 Hdist.
rewrite /dist_Rp in Hrel Hdist.
elim (normM_ex p (dif_m  (f (v0)) (f v)) ) => k Hnm.
have H1: forall x, cf x >R0.
move => x.
apply Rge_gt_trans with (min_del cf).
apply Rle_ge; apply min_min.
rewrite /Rgt.
elim Hdist => Hd1 Hd2.
fourier.
have H2:  forall x, R0 < norm (dif_v v0 v) < cf x.
move => x.
 elim Hdist => Hd1 Hd2.
split.
by done.
apply Rlt_le_trans with (min_del cf).
by done.
by apply min_min.
have H3:= (Hrel k).
elim H3 => H4 H5.
have H6:= (H5 v0 (H2 k)).
rewrite Hnm /sum_abs.
have H7: forall x , Rabs ((dif_m (f v0) (f v) k) x)  < eps/INR p.
move =>x.
elim (norm_rew Hp (dif_v (\vec_(i0)(f v0 k i0)) (\vec_(i0) (f v k i0)))) => j Hrelv.
apply Rle_lt_trans with ( norm (dif_v (\vec_(i0)(f v0 k i0)) (\vec_(i0) (f v k i0)))).
rewrite !mxK.
have H8: f v0 k x - f v k x = dif_v (\vec_(i0) (f v0 k i0)) (\vec_(i0) (f v k i0)) x 
by rewrite  !vxK.
rewrite H8.
apply norm_max.
exact H6.
apply: (ineq_sum_el Hp).
rewrite /ssralg.Ring.zero //=; fourier.
assumption.
Qed.


Lemma lema15:
forall (f: (Rvec p)->MR_(p)) (v:Rvec p) (Un:nat->Rvec p)  ,
(forall eps,R0<eps-> exists del, R0<del /\forall v0, R0< dist_Rp v0 v < del -> norm_m (dif_m (f v0) (f v))<eps) ->
conv_Rp Un v ->
(forall eps, R0<eps ->exists N, forall n, (N<=n)%nat->  norm_m (dif_m (f (Un n)) (f v))<eps).
Proof.
move =>f v Un H H0 eps0 He. 
destruct (H eps0 He) as [delta [H1 H2]] .
destruct (H0 delta H1) as [N H3] .
exists N.
intros n Hn.
elim (dist_pos  (Un n) v) => H4.
apply (H2 (Un n)).
split; first by done.
exact (H3 n Hn).
have Heq: Un n = v.
by rewrite -(dist_refl Hp).
rewrite Heq.
have H5: dif_m (f v) (f v) = \0.
 rewrite -matrixP /dif_m  =>i j; rewrite !mxK /ssralg.Ring.zero //=; ring.
by rewrite H5 zeroM_0.
Qed.

Lemma cont_lim_0:
forall (f:(Rvec p)->MR_(p)) (yn xn:nat->Rvec p) (y:Rvec p),
(forall v k, cont_Rp (fun v0 :Rvec p =>(\vec_(j) (f v0 k j))) v)->
(conv_Rp xn (vect0 p) )->
(conv_Rp yn y)->
(conv_Rp (fun i:nat=>mult_mv (f (yn i)) (xn i)) (vect0 p)).
Proof.
move => f yn xn y Hcont Hcvx Hcvy.
have H4: norm_m (f y) = R0 \/ norm_m (f y) <> R0.
elim (Rtotal_order (norm_m (f y)) R0)=> H5; auto with real.
elim H4 => H5.
move => eps He.
have Hsqrt: R0< sqrt eps.
by apply sqrt_lt_R0.
elim (Hcvx (sqrt eps) Hsqrt) => Nx Hrel2.
have Hco: (forall eps, R0<eps ->exists N, forall n, (N<=n)%nat->  norm_m (dif_m (f (yn n)) (f y))<eps).
apply lema15.
apply cont_norm_matr.
exact (Hcont y).
exact Hcvy.
elim (Hco (sqrt eps) Hsqrt) => Ny Hrel.
exists (max Nx Ny).
move => n Hnmax.
rewrite /dist_Rp dif_vect0.
have H1: mult_mv (f (yn n)) (xn n) = dif_v (mult_mv (f (yn n)) (xn n)) (mult_mv (f y) (vect0 p)).
by rewrite mult_v0 -ext_eq /dif_v / vect0 / mult_mv =>i; rewrite !vxK //=;  simp_R.
rewrite H1.
have H3: dif_v (mult_mv (f (yn n)) (xn n)) (mult_mv (f y) (vect0 p)) = 
   add_v (dif_v (mult_mv (f (yn n)) (xn n))( mult_mv (f y) (xn n))) (dif_v (mult_mv (f y) (xn n))(mult_mv (f y) (vect0 p))).
by rewrite -ext_eq /add_v /dif_v => i; rewrite !vxK; simp_R.
rewrite H3.
apply Rle_lt_trans with (norm (dif_v (mult_mv (f (yn n)) (xn n))( mult_mv (f y) (xn n)))+norm (dif_v (mult_mv (f y) (xn n))(mult_mv (f y) (vect0 p)))).
apply (norm_tri Hp).
rewrite distr_opp_mv distr_opp_vm.
apply Rle_lt_trans with (norm_m (dif_m (f (yn n)) (f y))* norm (xn n) + norm_m (f y)*norm (dif_v (xn n) (vect0 p))).
apply Rplus_le_compat; apply (ineq_norm_mv p ).
rewrite H5 Rmult_0_l Rplus_0_r. 
have H6:  sqrt eps* sqrt eps = eps.
 apply sqrt_sqrt; fourier.
rewrite -H6.
apply Rmult_ge_0_gt_0_lt_compat.
have H7:= norm_pos (xn n); fourier.
fourier.
apply Hrel.
apply le_trans with (max Nx Ny).
apply le_max_r.
done.
rewrite /dist_Rp in Hrel2.
replace (xn n) with (dif_v (xn n) (vect0 p)).
apply Hrel2.
apply le_trans with (max Nx Ny).
apply le_max_l.
done.
by rewrite -ext_eq /dif_v /vect0 => i; rewrite !vxK; ring_simplify.
move => eps He.
have Hpos: R0 < norm_m (f y).
have Hp1:= normM_pos p (f y).
elim Hp1 => Hp2; [done|by elim H5].
have Hpos1: R0<eps/(2*norm_m (f y)).
rewrite /Rdiv Rinv_mult_distr. 
rewrite -Rmult_assoc.
repeat apply Rmult_lt_0_compat.
fourier.
fourier.
auto with real.
 apply Rgt_not_eq; fourier.
done.
have Hco: (forall eps, R0<eps ->exists N, forall n, (N<=n)%nat->  norm_m (dif_m (f (yn n)) (f y))<eps).
apply lema15.
apply cont_norm_matr.
exact (Hcont y).
exact Hcvy.
elim (Hco (norm_m (f y)) Hpos) => Nx Hrel2.
elim (Hcvx (eps/(2*norm_m (f y))) Hpos1) => Ny Hrel.
exists (max Nx Ny).
move => n Hnmax.
rewrite /dist_Rp dif_vect0.
have H1: mult_mv (f (yn n)) (xn n) = dif_v (mult_mv (f (yn n)) (xn n)) (mult_mv (f y) (vect0 p)).
rewrite mult_v0 -ext_eq /dif_v / vect0 / mult_mv =>i; rewrite !vxK; by simp_R.
rewrite H1.
have H3: dif_v (mult_mv (f (yn n)) (xn n)) (mult_mv (f y) (vect0 p)) = add_v (dif_v (mult_mv (f (yn n)) (xn n))( mult_mv (f y) (xn n))) (dif_v (mult_mv (f y) (xn n))(mult_mv (f y) (vect0 p))).
rewrite -ext_eq /add_v /dif_v => i; rewrite !vxK; by ring_simplify.
rewrite H3.
apply Rle_lt_trans with (norm (dif_v (mult_mv (f (yn n)) (xn n))( mult_mv (f y) (xn n)))+norm (dif_v (mult_mv (f y) (xn n))(mult_mv (f y) (vect0 p)))).
apply (norm_tri Hp).
rewrite distr_opp_mv distr_opp_vm.
apply Rle_lt_trans with (norm_m (dif_m (f (yn n)) (f y))* norm (xn n) + norm_m (f y)*norm (dif_v (xn n) (vect0 p))).
apply Rplus_le_compat; apply (ineq_norm_mv p ).

have H6: norm (dif_v (xn n) (vect0 p)) = R0 \/ norm (dif_v (xn n) (vect0 p)) <> R0.
elim (Rtotal_order (norm (dif_v (xn n) (vect0 p))) R0)=> H7; auto with real.
elim H6 => H7.
pattern (xn n) at 1;replace (xn n) with (dif_v (xn n) (vect0 p)).
by rewrite H7 Rmult_0_r Rmult_0_r Rplus_0_l.
rewrite -ext_eq /dif_v /vect0 => i; rewrite !vxK; by ring_simplify.
pattern (xn n) at 1;replace (xn n) with (dif_v (xn n) (vect0 p)).
rewrite -Rmult_plus_distr_r.
replace eps with ((norm_m (f y) +norm_m (f y) )*(eps / (2*norm_m (f y)))); [|field].
apply Rmult_gt_0_lt_compat.
elim (norm_pos  (dif_v (xn n) (vect0 p))) => H8;[done|by elim H7].
fourier.
apply  Rplus_lt_compat_r .
apply Hrel2.
apply le_trans with (max Nx Ny).
apply le_max_l.
done.
rewrite /dist_Rp in Hrel.
apply Hrel.
apply le_trans with (max Nx Ny).
apply le_max_r.
done.
done.
rewrite -ext_eq /dif_v /vect0 => i ; rewrite !vxK; by ring_simplify.
Qed.


Lemma lim_prod_minus:
forall (f:Rvec p->MR_(p)) (yn:nat->Rvec p) (y:Rvec p),
(forall v k, cont_Rp (fun v0 :Rvec p =>(\vec_(j) (f v0 k j))) v)->
(conv_Rp yn y)->
(conv_Rp (fun i:nat=>mult_mv (f (yn i)) (dif_v (yn i)  (yn (S i)))) (vect0 p)).
Proof.
move=> f yn y Hcont Hcv.
apply cont_lim_0 with y; try done.
apply lim_minus_0 with y; done.
Qed.

End funp.

Implicit Arguments limit_Rp_R [p].
Implicit Arguments limit_Rp [p].
Implicit Arguments limit_R_Rp [p].
Implicit Arguments cont_Rp_R [p].
Implicit Arguments cont_Rp [p].
Implicit Arguments cont_R_Rp [p].