Require Import ssreflect eqtype fintype ssrbig matrix.
Require Import structRp min_max vectors seqRp funRp matrixNorm matrixRp deriv.
Require Import Fourier  Reals.

Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

Section Kantorovitch.

Variable p: pos_nat.
Lemma Hp: (O<p)%nat.
Proof.
have Hp1:= pos_natP p.
move/ssrnat.leP:Hp1=>Hp1.
omega.
Qed.
 
Notation "v1 +^ v2" := (add_v v1 v2) (at level 50).
Notation "a *^ v" := (mult_sv a v) (at level 40).
Notation "v1 -^ v2" := (dif_v v1 v2) (at level 50).
Notation "-^ v" := (opp_v v) (at level 35).
Notation "'MR_' p" := (matrix structRp.real_field p p)(at level 10).
Notation "A -m B" := (dif_m A B) (at level 50, format "A  -m  B").
Notation "A ^-1m" := (invmx  A) (at level 10, format "A ^-1m") .

Variable a0:R.
Variable b0:R.
Variable c:R.
Variable keps:R.
Variable f:(Rvec p)->Rvec p.
Variable X0:Rvec p.

Fixpoint An (n:nat) {struct n} : R :=
match n with
|0=>a0
| S n => 2*(An n)
end.

Fixpoint Bn (n:nat) : R:=
match n with
|0=>b0
|S n => INR p*(An n)*(Bn n)*(Bn n)*c
end.

Fixpoint En (n:nat): R :=
match n with
|0 => keps
|S n => (En n)/2
end.

Hypothesis Hkepos: 0<keps.
(*
Hypothesis H_a0:0<a0.
Hypothesis H_b0:0<b0.
Hypothesis H_c:0<c.
*)

Hypothesis A0_b0_c: 2*(INR p)*a0*b0*c<=1.

Definition c_disc_Rp (x:Rvec p) (delta : R)(y:Rvec p) : Prop :=
0<delta/\(dist_Rp x y)<=delta.


Hypothesis Hcontf:
forall v, cont_Rp  f v.
(* prove that a diff fun is continuous!*)

Hypothesis pr: forall (x : Rvec p) (x0 : I_(p)), dbl_pt p  f x0 x. (* once derivable*)
Hypothesis pr2: forall (i j:I_(p)) (a:Rvec p), dbl_pt_2 p  f i j a (fun x => pr x i) . (*twice der*) 
Hypothesis der_cont1: forall a i, cont_Rp  (fun x => der_p p  f i x (pr x i)) a. (*cont 1st der*)
Hypothesis der_cont2: forall a i j, cont_Rp  (fun v => der_p_2 p  f i j v (fun x => pr x i) (pr2 i j v)) a .
(*cont 2nd der*)

(*this is unnecessary... but it simplifies stuff;*)
Hypothesis der_bo: forall (v0:Rvec p) (i j : I_(p)),
\sum_( k < p) Rabs (der_p_2 p  f j k v0 (fun v => pr v j) (pr2 j k v0) i) <=c.

(*
Hypothesis pr: forall a i, dbl_pt p Hp f i a.
*)

Definition f' (a:Rvec p):=
Jac p  f a (pr a).

Hypothesis Hinv0 : \det (f' X0) <> 0.
Hypothesis Ha: norm_m ((f' X0)^-1m) <= a0.
Hypothesis Hb: norm (mult_mv ( (f' X0)^-1m) (f X0)) <= b0.
Hypothesis Hb0: b0<=keps*/2.



Lemma Hder_bo:
forall x y, norm_m (dif_m (f' x) (f' y)) <=c*(INR p)*norm (dif_v x y).
Proof.
move=> x y.

have H3: forall f g i a pr1 pr2, (forall t , f t = g t) -> dp_1 p f i a pr1 = dp_1 p g i a pr2.
move=> h g ind av prh prg Hhg.
rewrite /dp_1.
case prh => lh Hlh; rewrite //=.
case prg => lg Hlg; rewrite //=.
rewrite /part_deriv_pt_1 in Hlh Hlg.
have H1: limit_R (fun t : Base R_met => (g (av +^ t *^ base_v ind) - g av) / t)
        0 lh.
red; apply echiv_lim_f1 with (fun t : Base R_met => ((h (av +^ t *^ base_v ind) - h av) / t):Base R_met);
last by done .
by move=> t;  rewrite !Hhg.
by apply uniq_lim_R with 
(fun t : Base R_met => ((g (av +^ t *^ base_v ind) - g av) / t) : Base R_met) R0.
rewrite /f' /Jac.
have Heq1: x = (y+^(x-^y)).
 rewrite -ext_eq => i; rewrite !vxK; simp_R.
by rewrite /Rminus  Rplus_comm Rplus_assoc Rplus_opp_l Rplus_0_r.
rewrite {1 2}Heq1.
rewrite Rmult_assoc Rmult_comm.
have Hl1:= (lemme1 p  (fun i j y=> der_p p  f j y (pr y j) i ) y (x-^y) c).
have pr0 : forall (x x0 x1 : I_(p)) (x2 : Rvec p),
              dbl_pt_1 p (fun y : Rvec p => der_p p  f x0 y (pr y x0) x) x1
                x2.
move=>  k j i v.
have Hedp:= eq_der_p p  f j v (pr v j) k .
have Heq2: forall v, der_p p  f j v (pr v j) k =
       dp_1 p (fun v : Rvec p => f v k) j v
         (echiv_der_p2 p  f j v (pr v j) k).
 by move=> v0; apply eq_der_p.
have hdbl2: dbl_pt_1 p (fun v => dp_1 p (fun v0 : Rvec p => f v0 k) j v
         (echiv_der_p2 p  f j v (pr v j) k)) i v.
apply dbl_2_impl_dp_dp_1.

have H31: dbl_pt_2 p  f j i v (fun v => pr v j).
apply pr2.
have H4:= echiv_der_p2_gen p  f j i v (fun v=> pr v j) H31 k.
elim H4 => dp2 Hdp2; exists dp2.
red in Hdp2 |-*; red in Hdp2 |-*; red in Hdp2 |-*.
apply echiv_lim_f1   with (fun t : Base R_met =>
          (dp_1 p (fun v : Rvec p => f v k) j (v +^ t *^ base_v i)
             (echiv_der_p3 p  f j k (fun v : Rvec p => pr v j)
                (v +^ t *^ base_v i)) -
           dp_1 p (fun v : Rvec p => f v k) j v
             (echiv_der_p3 p  f j k (fun v : Rvec p => pr v j) v)) / t); last by done.
move => t.
have heq1: dp_1 p (fun v0 : Rvec p => f v0 k) j (v +^ t *^ base_v i)
   (echiv_der_p3 p  f j k (fun v0 : Rvec p => pr v0 j) (v +^ t *^ base_v i)) =
(dp_1 p (fun v0 : Rvec p => f v0 k) j (v +^ t *^ base_v i)
   (echiv_der_p2 p  f j (v +^ t *^ base_v i) (pr (v +^ t *^ base_v i) j) k)).
apply dp_1_pr_nu.
have heq2: dp_1 p (fun v0 : Rvec p => f v0 k) j v
   (echiv_der_p3 p  f j k (fun v0 : Rvec p => pr v0 j) v)  =
dp_1 p (fun v0 : Rvec p => f v0 k) j v (echiv_der_p2 p  f j v (pr v j) k) .
apply dp_1_pr_nu.
by rewrite heq1 heq2.
elim hdbl2 => dp2 Hdp2.
exists dp2.
red in Hdp2 |-*; red in Hdp2 |-*.
apply echiv_lim_f1 with  (fun t : Base R_met =>
          (dp_1 p (fun v0 : Rvec p => f v0 k) j (v +^ t *^ base_v i)
             (echiv_der_p2 p  f j (v +^ t *^ base_v i)
                (pr (v +^ t *^ base_v i) j) k) -
           dp_1 p (fun v0 : Rvec p => f v0 k) j v
             (echiv_der_p2 p  f j v (pr v j) k)) / t); last by done.
move => t.
have heq1: (dp_1 p (fun v0 : Rvec p => f v0 k) j (v +^ t *^ base_v i)
   (echiv_der_p2 p  f j (v +^ t *^ base_v i) (pr (v +^ t *^ base_v i) j) k))=
(der_p p  f j (v +^ t *^ base_v i) (pr (v +^ t *^ base_v i) j) k).
by rewrite Heq2.
have heq2:  dp_1 p (fun v0 : Rvec p => f v0 k) j v (echiv_der_p2 p  f j v (pr v j) k)=
der_p p  f j v (pr v j) k.
by rewrite Heq2.
by rewrite heq1 heq2.
have Hl2:= Hl1 pr0.
have Hcont1: forall (i j i0 : I_(p)) (v0 : Base (Rp_met p )),
       cont_Rp_R 
         (fun x : Rvec p =>
          dp_1 p (fun y : Rvec p => der_p p  f j y (pr y j) i) i0 x
            (pr0 i j i0 x)) v0.
move=> j i k v.

apply cont_eq_fun with (fun v => (der_p_2 p  f i k v (fun x => pr x i) (pr2 i k v) j)).
move=> x0.
rewrite dp_12_der_p2_eq dp_2_impl_dp_1.

apply H3.
move=> w.
rewrite eq_der_p.
apply H3.
move=> t; done.
apply: rel_cont_Rp_R_Rp.
apply der_cont2.
have Hl3:= Hl2 Hcont1.
have Hbound:  forall (x : Rvec p) (i j : I_(p)),
\sum_(k<p) Rabs  (dp_1 p (fun y : Rvec p => der_p p  f j y (pr y j) i) k x (pr0 i j k x)) <= c.
move => v0 i j.
apply Rle_trans with (\sum_( k < p) Rabs (der_p_2 p  f j k v0 (fun v => pr v j) (pr2 j k v0) i) ).
right.
apply: eq_bigr => k _.
f_equal.
rewrite dp_12_der_p2_eq dp_2_impl_dp_1.
apply H3.
move=> t.
rewrite eq_der_p.
apply H3.
move=> w; done.
apply der_bo.
apply (Hl3 Hbound).
Qed.




Fixpoint Xn (n:nat): Rvec p:=
match n with 
|0 => X0
|S n =>dif_v (Xn n) (mult_mv ((f' (Xn n))^-1m) (f (Xn n)))
end . 



Lemma ineq12:
forall (x y:Rvec p),
norm (dif_v (f x) (add_v (f y)  (mult_mv  (f' y) (dif_v x y))))<=1/2*INR p*Rsqr (norm (dif_v x y))*c.
Proof.
move => x y.
have Heqy: x = y +^ (x -^ y).
rewrite -ext_eq => i; rewrite !vxK; simp_R.
by rewrite /Rminus  Rplus_comm Rplus_assoc Rplus_opp_l Rplus_0_r.
rewrite {1}Heqy.
have Heq: (f (y +^ (x -^ y)) -^ (f y +^ mult_mv (f' y) (x -^ y)))=
  ((f (y +^ (x -^ y)) -^ f y) -^ mult_mv (f' y) (x -^ y)).
by rewrite -ext_eq => i; rewrite !vxK; ring_simplify.
rewrite Heq.
by rewrite /f'; apply lemme2 with pr2.
Qed.

Lemma keps_pos:
forall n, 0<En n.
Proof.
induction n.
simpl; auto.
simpl; fourier.
Qed.

Lemma inc_neigh_Rp:
forall x y r, 0<r -> dist_Rp x y <= r*/2 ->
(forall v, c_disc_Rp y (r*/2) v -> c_disc_Rp x r v).
Proof.
move => x y r Hr Hdist v [He0 Hdisc].
unfold c_disc_Rp; split; first by done.
apply Rle_trans with (dist_Rp x y + dist_Rp y v).
 apply vectors.dist_tri; exact Hp.
rewrite -(eps2 r); fourier.
Qed.

Lemma inc_neigh_ech:
forall (x y:Rvec p) r, 0<r -> dist_Rp x y <= r*/2 ->
(forall v, dist_Rp y v<=(r*/2)  -> dist_Rp x v <=r).
Proof.
move => x y r Hr H1 v H2.
apply Rle_trans with (dist_Rp x y + dist_Rp y v).
by apply vectors.dist_tri; exact Hp.
rewrite -(eps2 r); fourier.
Qed.


Lemma Rinv_ineq:
forall a b,0<=a->0<=b->
1<=a*b -> 1/b<=a.
Proof.
move =>a b H1 H2 H3.
unfold Rdiv.
have H4: 0<b.
have H5: 0<a*b.
apply Rlt_le_trans with 1; fourier.
have Hco: ~ (b<=0).
unfold not => Hco1.
have Hco4: b=0. 
by apply Rle_antisym.
have Hco2: 0=a*b. by rewrite Hco4; ring.
have Hco3: ~(0=a*b).
by apply Rlt_not_eq.
contradiction.
elim (Rle_or_lt b 0)=>H.
contradiction.
done.
replace a with (a*b*/b); [|field].
apply Rmult_le_compat_r.
unfold Rle; left.
auto with real.
assumption.
apply Rgt_not_eq.
fourier.
Qed.

Lemma norm_m_pos:
forall x, 0<= @norm_m p x.
Proof.
move => x.
elim (normM_ex p  x) => k Heq.
rewrite Heq.
apply (sum_abs_pos p ).
Qed.

Lemma norm_m_inv:
forall a, (a^-1m) *m a = \1_(p) ->
 1/(norm_m a) <= norm_m (a^-1m).
Proof.
move => a H.
have H2: 1 <= norm_m ( a^-1m)*norm_m a.
have H1: norm_m \1_(p) = 1.
apply (norm_unitM p ).
rewrite -H1 -H.
apply (ineq_norm_m p ).
apply Rinv_ineq.
apply norm_m_pos.
apply norm_m_pos.
assumption.
Qed.

Lemma a0_ge_0:
0<=a0.
Proof.
apply Rle_trans with (norm_m ((f' X0)^-1m));[
apply norm_m_pos|done].
Qed.

Lemma b0_ge_0:
0<=b0.
Proof.
apply Rle_trans with (norm (mult_mv ( (f' X0)^-1m) (f X0))).
 apply norm_pos; exact Hp.
done.
Qed.

Lemma c_ge_0:
0<=c.
Proof.
apply Rle_trans with (
\sum_( k < p) Rabs (der_p_2 p  f (I0 p Hp) k X0 (fun v => pr v (I0 p Hp)) (pr2 (I0 p Hp) k X0) (I0 p Hp))).
apply Rle_trans  with 
(Rabs (\sum_( k < p) (der_p_2 p  f (I0 p Hp) k X0 (fun v => pr v (I0 p Hp)) (pr2 (I0 p Hp) k X0) (I0 p Hp)))).
apply Rabs_pos.
by apply: abs_sum_le_f.
done.
Qed.



Lemma An_gt_0:
forall n, 0<=An n.
Proof.
induction n.
simpl; auto.
apply a0_ge_0.
simpl; apply Rmult_le_pos; [fourier|auto].
Qed.

Lemma An_bn_c0:
forall (n:nat) , 0<=2*INR p*(An n)*(Bn n)*c.
Proof.
induction n.
rewrite //=.
 repeat apply Rmult_le_pos. 
 fourier.
 apply pos_INR.
apply a0_ge_0.
apply b0_ge_0.
apply c_ge_0.
simpl.
replace (2 * INR p*(2 * An n) * (INR p*An n * Bn n * Bn n * c) * c) with ((2*INR p*An n*Bn n * c)*(2*INR p*An n*Bn n * c));
 [idtac | ring].
fold (Rsqr (2*INR p*An n*Bn n*c)).
auto with real.
Qed.

Lemma An_bn_c1:
forall (n:nat) , 2*(INR p)*(An n)*(Bn n)*c<=1.
Proof.
induction n.
simpl; auto.
simpl.
replace (2 * INR p*(2 * An n) * (INR p*An n * Bn n * Bn n * c) * c) with ((2*INR p*An n*Bn n * c)*(2*INR p*An n*Bn n * c));
 [idtac | ring].
fold (Rsqr (2*INR p*An n*Bn n*c)).
pattern 1 at 3;rewrite <- Rsqr_1.
apply Rsqr_incr_1.
done.
apply An_bn_c0.
fourier.
Qed.

Lemma Bn_gt_0:
forall n, 0<=Bn n.
Proof.
induction n.
simpl; auto.
apply b0_ge_0.
simpl; repeat apply Rmult_le_pos; auto.
by apply pos_INR.
by apply An_gt_0.
apply c_ge_0.
Qed.

Lemma Bn_en_le:
forall n,
Bn ( n) <= En ( n) * / 2.
Proof.
induction n; rewrite //=.
replace (INR p * An n * Bn n * Bn n * c) with (Bn n*(2*INR p*An n* Bn n*c)*/2); [|field].
apply Rmult_le_compat_r.
fourier.
replace (En n/2) with (En n/2*1);[|field].
apply Rmult_le_compat.
 apply Bn_gt_0. 
 apply An_bn_c0.
done.
apply An_bn_c1.
Qed.

Lemma bad_four:
forall r,0<=r -> r<=1/2 ->
1/(1-r) <=2.
Proof.
move => r H1 H2.
have H3:1/2 <=  1-r .
fourier.
apply Rinv_ineq.
fourier.
fourier.
pattern 1 at 1;replace 1 with (2*(1/2)); [|field].
apply Rmult_le_compat; fourier.
Qed.


Lemma about_seq:
forall n, 
(n<>0%nat-> forall v, 
   (dist_Rp (Xn n) v <= (En n))->
   (dist_Rp (Xn (pred n)) v <= (En (pred n)))) /\
\det (f' (Xn n)) <>0/\
norm_m ( (f' (Xn n))^-1m) <= An n/\
norm (mult_mv ((f' (Xn n))^-1m) (f (Xn n)))<= Bn n/\
Bn n<= (En n)*/2.
Proof.
induction n.
rewrite //=.
elim IHn => [IHdisc [IHinv [IHa [IHb1 IHb2]]]].
have  Hinvx0:  \det (f' (Xn n)) != 0.
case X: (\det (f' (Xn n)) == 0); rewrite //=.
move/eqP: X =>X.
have H3 := IHinv X.
done.
split.
move => Hdif v Hdisc.
rewrite //= .
apply inc_neigh_ech with (Xn (S n)).
 by apply keps_pos.
rewrite /dist_Rp.
have Heq: dif_v (Xn n) (Xn (S n)) = mult_mv ((f' (Xn n))^-1m) (f (Xn n)).
by apply/ext_eq =>i; rewrite vxK //= vxK; ring_simplify.
rewrite Heq; by apply Rle_trans with (Bn n).
by rewrite //En in Hdisc.
have Hinv: (f' (Xn n))^-1m *m (f' (Xn n)) = \1.
apply: mulmxV.
done.
have Ha3: norm_m (dif_m \1 (( (f' (Xn n))^-1m) *m (f' (Xn (S n))))) <= 1/2.
rewrite -Hinv.
rewrite distr_m_dif.
apply Rle_trans with ((norm_m ((f' (Xn n))^-1m))*
 norm_m ((dif_m (f' (Xn n)) ((f' (Xn (S n))))))).
apply (ineq_norm_m p ).
apply Rle_trans with ((An n)*norm_m (dif_m (f' (Xn n)) (f' (Xn (S n))))).
apply Rmult_le_compat.
elim (normM_ex p  ((f' (Xn n))^-1m)) => k Heq.
rewrite Heq.
apply (sum_abs_pos p ).
elim (normM_ex p  (dif_m (f' (Xn n)) (f' (Xn (S n))))) => k Heq.
rewrite Heq.
apply (sum_abs_pos p ).
done.
by unfold Rle; right.
apply Rle_trans with ((An n)*((INR p)*(Bn n)*c)).
apply Rmult_le_compat_l.
apply An_gt_0.
have Ha2: norm (dif_v (Xn n) (Xn (S n))) <= Bn n.
have Heq2: dif_v (Xn n) (Xn (S n)) = mult_mv ( (f' (Xn n))^-1m) (f (Xn n)).
rewrite //=.
by apply/ext_eq =>i; rewrite vxK //= vxK; ring_simplify.
by rewrite Heq2.
replace (INR p*Bn n*c) with (c*INR p*Bn n);[| ring].
apply Rle_trans with (c*INR p* norm (dif_v (Xn n) (Xn (S n)))).
apply Hder_bo.
apply Rmult_le_compat_l.
apply Rmult_le_pos.
exact c_ge_0.
 apply pos_INR.
exact Ha2.
replace  (An n * (INR p * Bn n * c)) with (1/2*(2*INR p*An n*Bn n*c)); [|field].
pattern (1/2) at 2; replace (1/2) with (1/2*1); [|field].
apply Rmult_le_compat_l.
 fourier.
apply An_bn_c1.
have Ha2: \det ((f' (Xn n))^-1m *m (f' (Xn (S n)))) <> 0.
apply matr_inv_norm.
apply Rle_lt_trans with (1/2);[done|].
rewrite /ssralg.Ring.one //=. fourier.
rewrite determinantM in Ha2.
have Hinvs: \det (f' (Xn (S n))) <> 0.
rewrite /ssralg.Ring.mul in Ha2.
by elim (Rmult_neq_0_reg  _ _ Ha2) .
split; first by done.
have Ha1: (f' (Xn (S n)))^-1m = (( (f' (Xn n))^-1m *m (f' (Xn (S n))))^-1m)*m ((f' (Xn n))^-1m).
 rewrite invmx_mul .
rewrite -  mulmxA.
rewrite invmxK.
rewrite mulVmx .
rewrite  mulmx1; done.
done.
done.
apply neq0_Dmx; done.
case X: (\det (f' (Xn (S n))) == 0); rewrite //=.
move/eqP: X =>X.
have H3 := Hinvs X.
done.
have Ha4:  norm_m ((f' (Xn (S n))^-1m)) <= An (S n).
rewrite Ha1.
apply Rle_trans with (norm_m (((f' (Xn n))^-1m *m f' (Xn (S n)))^-1m) * norm_m ((f' (Xn n))^-1m)).
apply (ineq_norm_m p ).
simpl An.
apply Rmult_le_compat .
apply norm_m_pos.
apply norm_m_pos.
replace ( ( (f' (Xn n))^-1m) *m (f' (Xn (S n)))) with 
       (dif_m \1 (dif_m \1 (( (f' (Xn n))^-1m) *m (f' (Xn (S n)))))).
apply Rle_trans with (1/(1-( norm_m (dif_m \1 ( ((f' (Xn n))^-1m) *m (f' (Xn (S n)))))))).
apply norm_inv_ser.
apply Rle_lt_trans with (1/2);[done|].
rewrite /ssralg.Ring.one //=. fourier.
apply bad_four.
apply norm_m_pos.
done.
rewrite - matrixP /dif_m => i j; rewrite !mxK; by ring_simplify. 
by elim IHn => [_ [_ [H _]]].
split; first by done.
split.
simpl Bn.
apply Rle_trans with (norm_m ( (f' (Xn (S n)))^-1m)* norm (f (Xn (S n)))).
apply (ineq_norm_mv p ).
replace (INR p * An n * Bn n * Bn n * c) with (2*An n*(1/2*INR p* Bn n*Bn n*c)); [|field].
apply Rmult_le_compat.
apply norm_m_pos.
apply (norm_pos ).
by simpl An in Ha4.
have Hxn: f (Xn (S n)) = (dif_v (f (Xn (S n))) (add_v (f (Xn n))  (mult_mv (f'(Xn  n )) (dif_v (Xn (S n)) (Xn n))))).
have Hxna: (add_v (f (Xn n))  (mult_mv (f'(Xn  n )) (dif_v (Xn (S n)) (Xn n)))) = vect0 p.
simpl Xn.
replace (dif_v (dif_v (Xn n) (mult_mv ( (f' (Xn n))^-1m) (f (Xn n)))) (Xn n))
 with (mult_mv ( (f' (Xn n))^-1m) (opp_v (f (Xn n)))).
rewrite -ext_eq /dif_v /vect0=>i; rewrite vxK.
rewrite mult_mv_assoc .
rewrite mulVmx .
rewrite unitM_el_neu.
by rewrite  !vxK; simp_R.
done.
by rewrite opp_distr -ext_eq /dif_v /opp_v  => i ; rewrite !vxK; simp_R.
rewrite Hxna -ext_eq /vect0 /dif_v =>i ; rewrite !vxK; simp_R.
auto with real.
rewrite Hxn.
apply Rle_trans with (1/2 *INR p* (Rsqr (norm (dif_v (Xn (S n)) (Xn n))))*c).
apply ineq12.
apply Rmult_le_compat_r.
apply c_ge_0.
replace (1/2*INR p*Bn n*Bn n) with ((1/2*INR p)*(Bn n*Bn n)); [|ring].
apply Rmult_le_compat_l.
unfold Rle; left; apply Rmult_lt_0_compat.
fourier.
by apply lt_INR_0; exact Hp.
fold (Rsqr (Bn n)).
fold (dist_Rp (Xn (S n)) (Xn n)).
rewrite vectors.dist_sym.
apply Rsqr_incr_1.
rewrite /dist_Rp.
replace (dif_v (Xn  n) (Xn (S n))) with (mult_mv ( (f' (Xn n))^-1m) (f (Xn n))).
by elim IHn => [_[_[_[H _]]]].
by rewrite  //= -ext_eq /dif_v =>i; rewrite !vxK; simp_R.
apply (vectors.dist_pos ).
apply Bn_gt_0.
apply Bn_en_le.
Qed.

Lemma Cauchy_crit_Rp2:
forall Un,
Cauchy_crit_Rp p Un <->
(forall eps0:R,
eps0>0->
exists N:nat,
(forall n p:nat,
(n>=N)%nat->dist_Rp (Un n) (Un (n+p)%nat) < eps0)).
Proof.
move =>Un.
split => H.
unfold Cauchy_crit_Rp in H.
move => eps0 H0.
elim (H eps0 H0); intros N Hchy.
exists N.
move => n p0 Hineq.
apply (Hchy n (n+p0)%nat).
assumption.
omega.
unfold Cauchy_crit_Rp.
move =>eps0 H0.
elim (H eps0 H0)=> N Hcrit.
exists N.
move => n m Hineq.
elim (le_or_lt n m); intros.
replace m with (n+(m-n))%nat;[| omega].
apply (Hcrit n (m-n)%nat).
assumption.
rewrite vectors.dist_sym.
replace n with (m+(n - m))%nat;[|omega].
apply (Hcrit m (n - m)%nat).
assumption.
Qed.

Lemma disc_incl_discRp:
( forall n : nat,  forall v : Rvec p,
         dist_Rp (Xn (S n)) v <= En (S n) -> dist_Rp (Xn n) v <= En  n) ->
(forall (n p0:nat), forall v, dist_Rp (Xn (n+p0)%nat) v <= En (n+p0)%nat -> dist_Rp (Xn n) v <= En  n).
Proof.
move => H n p0 v H1.
induction p0.
by replace (n+0)%nat with n in H1; [|omega].
replace (n+S p0)%nat with (S (n+p0)%nat) in H1;[|omega].
have H2: dist_Rp (Xn (n + p0)%nat) v <= En (n + p0)%nat.
apply (H (n+p0)%nat v H1).
exact (IHp0 H2).
Qed.

Lemma norm_eq:
forall v:Rvec p,
dist_Rp v v = 0.
Proof.
move => v.
rewrite /dist_Rp norm0_eq; [exact Hp|done].
Qed.

Lemma disc_incl_pct:
( forall n : nat,  forall v : Rvec p,
         dist_Rp (Xn (S n)) v <= En (S n) -> dist_Rp (Xn n) v <= En  n) ->
(forall n p0, dist_Rp (Xn n) (Xn (n+p0)%nat) <= En n).
Proof.
move => H1 n p0 .
have H0:= (disc_incl_discRp H1). 
have H3:= H0 n p0.
apply (H3 (Xn (n+p0)%nat)).
rewrite norm_eq.
unfold Rle; left; apply keps_pos.
Qed.

(*borrowed and new proofs to apply archimede's principle*)
Theorem INR_Zabs_nat_eq :
  forall (z:Z), (0 <= z)%Z -> (IZR z = INR(Zabs_nat z))%R.
unfold Zabs_nat.
intros z; case z.
simpl.
intros; auto.
unfold IZR.
intros; auto.
compute.
intros x H'; elim H'; auto.
Qed.

Lemma pow2_larger :
  forall n, (INR n <= 2^n)%R.
induction n.
simpl; fourier.
replace (S n) with (1+n)%nat.
rewrite pow_add; rewrite pow_1.
assert (Hlem: (forall a b, 1<= b -> a<=b -> 1+a<= 2*b)%R).
intros; fourier.
rewrite plus_INR; simpl (INR 1).
apply Hlem; auto.
apply pow_R1_Rle;fourier.
omega.
Qed.

Lemma pow_half_smaller :
 forall r, (0 < r) -> (1/(2^(Zabs_nat(up (1/r)))) <= r).
intros r Hpos.
elim (archimed (1/r)); intros Harc1 Harc2.
apply (Rmult_le_reg_l (2^(Zabs_nat (up (1/r))))).
apply pow_lt.
fourier.
replace (2 ^ Zabs_nat (up (1 / r)) * (1 / 2 ^ Zabs_nat (up (1 / r))))%R
 with 1%R.
apply Rle_trans with ((INR (Zabs_nat (up (1/r))))*r)%R.
rewrite <- INR_Zabs_nat_eq.
pattern 1%R at 1; replace 1%R with (1/r * r)%R;[| field].
apply Rmult_le_compat_r.
fourier.
destruct (archimed (1/r)).
unfold Rle; left.
unfold Rgt in H.
assumption.
by apply Rgt_not_eq.
apply le_IZR.
apply Rle_trans with (1/r)%R.
replace (1/r)%R with (/r)%R.
simpl IZR.
assert (0 < / r)%R.
apply Rinv_0_lt_compat; auto.
unfold Rle; left; assumption.
field.
apply Rgt_not_eq; auto.
destruct (archimed (1/r)).
unfold Rle; left.
unfold Rgt in H.
assumption.
apply Rmult_le_compat_r.
fourier.
apply pow2_larger.
field.
apply pow_nonzero.
discrR.
Qed.

Lemma archim:
forall v, 0 < v -> exists n:nat, (1/(2^n) <= v)%R.
intros v Hvpos.
exists (Zabs_nat (up (1/v)))%nat.
destruct (archimed (1/v)) as [Hb1 _].
apply pow_half_smaller.
assumption.
Qed.

Lemma archim_lt:
(forall v, 0 < v -> exists n:nat, 1/(2^n) <= v/2).
intros v Hvpos.
exists (Zabs_nat (up (1/v))+1)%nat.
destruct (archimed (1/v)) as [Hb1 _].
rewrite pow_add.
replace (1 / (2 ^ Zabs_nat (up (1 / v))*2^1))%R
 with ((1 / (2 ^ Zabs_nat (up (1 / v))))*/2)%R.
replace (v/2)%R with (v*/2)%R.
apply Rmult_le_compat_r.
assert (0 < /2)%R.
apply Rinv_0_lt_compat;auto.
fourier.
fourier.
apply pow_half_smaller.
fourier.
field.
replace (2^1)%R with 2%R.
field.
assert (2^Zabs_nat(up(1*/v)) <> 0)%R.
apply pow_nonzero;auto.
discrR.
simpl;assumption.
simpl;field.
Qed.

Lemma archim_lt2:
forall v, 0 < v -> exists n:nat, (1/(2^n) < v)%R.
intros v Hvpos.
assert (Hhalf: exists n:nat, 1/(2^n) <= v/2).
apply archim_lt; assumption.
elim Hhalf; intros n Hineq.
exists n.
assert (v/2<v).
fourier.
unfold Rle in Hineq; elim Hineq; intros.
apply Rlt_trans with (v/2); assumption.
rewrite H0.
assumption.
Qed.

Lemma pow2_diff0:
forall n:nat,
2^n>0.
induction n; simpl; fourier.
Qed.

Lemma archim2_lt:
forall v k, 0 < v ->0<k-> exists n:nat, (k/(2^n) < v)%R.
intros v k Hvpos Hkpos.
assert (exists n:nat, 1/2^n<v/k).
apply (archim_lt2 ).
unfold Rdiv.
apply Rlt_mult_inv_pos; assumption.
elim H; intros N Hprop.
exists N.
replace (k/2^N) with (k*(1/2^N));[| field].
replace v with (k*(v/k)).
apply (Rmult_lt_compat_l k (1/2^N) (v/k)).
fourier.
assumption.
field.
apply Rgt_not_eq.
fourier.
replace (/2^N) with (1/2^N);[| field].
trivial.
apply Rgt_not_eq.
apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.

Lemma pow2_monot:
forall (n :nat), keps/2^(S n)<=keps/2^n.
Proof.
induction n .
simpl.
fourier.
simpl in IHn |-*.
replace (keps/(2*(2*2^n))) with (1/2*(keps/(2*2^n))) ;[|field; auto;apply Rgt_not_eq; apply pow2_diff0].
pattern (keps/(2*2^n)) at 2;replace (keps/(2*2^n)) with (1/2*(keps/2^n));[| field; apply Rgt_not_eq; apply pow2_diff0].
apply (Rmult_le_compat_l (1/2) (keps/(2*2^n)) (keps/2^n)).
fourier.
assumption.
Qed.

Lemma pow2_monot2:
forall (n p:nat), 
keps/2^(n+p)<=keps/2^n.
intros n p0.
induction p0.
replace (n+0)%nat with n; [| omega].
unfold Rle; right; trivial.
replace (n+S p0 )%nat with (S (n+p0)%nat);[| omega].
assert (IH2: keps/2^S (n+p0)%nat<=keps/2^(n+p0)%nat).
apply pow2_monot.
apply Rle_trans with (keps/2^(n+p0)%nat); assumption.
Qed.

Lemma pow2_monot3:
forall (n N:nat), (n>=N)%nat->
keps/2^n<=keps/2^N.
intros.
replace n with (N+(n-N))%nat;[| omega].
apply (pow2_monot2 N (n-N)%nat).
Qed.

Lemma transf_en:
forall n:nat,
En n = keps/2^n.
Proof.
induction n.
simpl;field.
simpl En.
rewrite IHn.
simpl; field.
apply Rgt_not_eq.
apply pow2_diff0.
Qed.

Lemma An_bn_c12:
forall (n:nat) , INR p*(An n)*(Bn n)*Bn n*c<=1/2*Bn n.
Proof.
intros.
assert ( 2*INR p*(An n)*(Bn n)*c<=1).
apply An_bn_c1.
assert  (INR p*(An n)*(Bn n)*c<=1/2).
unfold Rdiv.
replace (INR p*An n*Bn n*c) with ((1/2)*(2*INR p*(An n)*(Bn n)*c));[| field].
replace (1*/2) with (1/2*1); [|field].
apply (Rmult_le_compat_l (1/2) (2*INR p* (An n)*(Bn n)*c) 1); fourier .
replace (INR p*(An n)*(Bn n)*Bn n*c) with (INR p*An n*Bn n*c*Bn n);[| ring].
apply Rmult_le_compat_r ; [|assumption].
apply Bn_gt_0;assumption.
Qed.

Lemma Bn_desc_2:
forall n:nat,
Bn n<=b0/2^n.
Proof.
induction n.
simpl; fourier.
simpl.
unfold Rdiv in IHn.
assert (1/2* Bn n<=b0*/(2*2^n)).
replace (b0*/(2*2^n)) with (1/2*(b0/2^n)).
apply Rmult_le_compat_l.
fourier.
assumption.
field.
apply Rgt_not_eq.
apply pow2_diff0.
apply Rle_trans with (1/2*Bn n).
apply An_bn_c12.
done.
Qed.

Lemma not_series:
forall n, dist_Rp (Xn n) (X0) <= 2*b0.
Proof.
have H1: forall n, dist_Rp (Xn (S n)) (Xn n) <= Bn n.
move => n.
rewrite vectors.dist_sym.
unfold dist_Rp.
replace (dif_v (Xn  n) (Xn (S n))) with (mult_mv ( (f' (Xn n))^-1m) (f (Xn n))).
by  elim (about_seq n) => [_[_[_[H _]]]].
by rewrite  //= -ext_eq /dif_v =>i; rewrite !vxK; simp_R. 
have H2: forall n, dist_Rp (Xn n) X0 <= 2*b0 - 2*b0/(2^n).
induction n.
rewrite //=  norm_eq.
replace (2*b0 - 2*b0/1) with 0; [|field; auto]; fourier. 
simpl (2^S n).
rewrite /dist_Rp.
replace (dif_v (Xn (S n)) X0) with (add_v (dif_v (Xn (S n)) (Xn n)) (dif_v (Xn n) X0)).
have H3: norm (dif_v (Xn (S n)) (Xn n)) <= b0/2^n.
apply Rle_trans with (Bn n).
apply (H1 n).
apply (Bn_desc_2).
apply Rle_trans with ( norm (dif_v (Xn (S n)) (Xn n)) + norm (dif_v (Xn n) X0)).
apply (norm_tri Hp).
replace (2*b0 - 2*b0/(2*2^n)) with (b0/2^n + (2*b0 - 2*b0/2^n)).
apply Rplus_le_compat; assumption.
unfold Rdiv.
field.
apply Rgt_not_eq.
apply pow2_diff0.
rewrite -ext_eq /add_v /dif_v => i; rewrite !vxK; simp_R; done.
move => n.
apply Rle_trans  with  (2 * b0 - 2 * b0 / 2 ^ n).
exact (H2 n).
replace (2*b0-2*b0/(2^n)) with (2*b0 + (-(2*b0*/(2^n)))); [|field;auto].
pattern (2*b0) at 3; replace (2*b0) with (2*b0+0) ;[| ring].
apply Rplus_le_compat_l.
replace 0 with (-0);[| ring].
apply Ropp_le_contravar.
apply Rle_mult_inv_pos.
apply Rmult_le_pos; [fourier|apply b0_ge_0].
apply pow2_diff0.
apply Rgt_not_eq; apply pow2_diff0.
Qed.


Lemma seq_lim_in_dom:
forall xn (l:Rvec p) r,
0<=r->
conv_Rp xn l ->
(forall n, dist_Rp (xn n) (xn 0%nat) <= r)->
dist_Rp l (xn 0%nat) <= r.
Proof.
move =>xn l r Hr Hcv Hin.
have Hco: ~  r<dist_Rp l (xn 0%nat) .
unfold not => Hno.
have H1:  0 < dist_Rp l (xn 0%nat)-r.
fourier.
rewrite /conv_Rp in Hcv.
have Hco1: exists N, forall n, (N<=n)%nat ->dist_Rp (xn n) l <dist_Rp l (xn 0%nat)-r .
apply (Hcv (dist_Rp l (xn 0%nat) - r) H1).
elim Hco1 => N Hlim.
have Hdist: dist_Rp (xn N) l < dist_Rp l (xn 0%nat) - r.
apply (Hlim N); omega.
have H3: r<r.
apply Rlt_le_trans with (dist_Rp (xn 0%nat) l - dist_Rp (xn N)l ).
rewrite vectors.dist_sym; fourier.
apply Rle_trans with (dist_Rp (dif_v (xn 0%nat) l) (dif_v (xn N) l)).
apply (dist_min_tri p Hp).
replace (dist_Rp (dif_v (xn 0%nat) l) (dif_v (xn N) l)) with (dist_Rp (xn 0%nat) (xn N)).
rewrite vectors.dist_sym; exact (Hin N).
rewrite /dist_Rp.
have H: (dif_v (xn 0%nat) (xn N))=(dif_v (dif_v (xn 0%nat) l) (dif_v (xn N) l)).
by rewrite -ext_eq /dif_v => i; rewrite !vxK; ring_simplify.
by rewrite H.
have H: ~(r<r); fourier.
elim (Rle_or_lt (dist_Rp l (xn 0%nat)) r) => H.
fourier.
contradiction.
Qed.

Lemma cont_eq_fun_Rp:
forall f g a,
(forall x, f x = g x)->
cont_Rp  f a ->
@cont_Rp p g a.
Proof.
by move=> h g a Hfg Hf;
rewrite /cont_Rp_R /cont_gen in Hf |-* =>eps He; elim (Hf eps He)=> alp [Ha1 Ha2]; 
exists alp; split; [done|move=>x Hx; rewrite -!Hfg; apply Ha2].
Qed.

Notation "\vec_ ( i ) E" :=
  (@vec_of_fun p (fun i  => E)) (at level 25).


Theorem kantoroRp_exists:
exists xs:Rvec p,
conv_Rp Xn xs /\ norm (dif_v xs X0) <= 2*b0 /\ f xs = vect0 p.
Proof.
have Hcc: Cauchy_crit_Rp p Xn.
have Hcc1: forall n,
  (n<>0%nat-> forall v, (dist_Rp (Xn n) v <= (En n))->
    (dist_Rp (Xn (pred n)) v <= (En (pred n)))).
move =>n.
by elim (about_seq n) => H.
rewrite Cauchy_crit_Rp2.
move => eps0 He.
have Hincl:forall n p, (dist_Rp (Xn n) (Xn (n+p)%nat) <= En n ).
apply disc_incl_pct.
intros n.
apply (Hcc1 (S n)).
omega.
assert (Harch: exists n:nat, (keps/(2^n) < eps0)).
apply archim2_lt; fourier.
elim Harch; intros N Heps.
exists N; intros n p0 H1.
assert (Hnp:dist_Rp (Xn n) (Xn (n+p0)%nat) <= En n).
apply (Hincl n p0).
rewrite transf_en in Hnp.
assert (keps/2^n<=keps/2^N).
apply (pow2_monot3 ); assumption.
apply Rle_lt_trans with (keps/2^N).
by apply Rle_trans with (keps/2^n).
done.
elim (Rp_complete p Hp Xn Hcc) => xs Hconv.
exists xs.
split.
done.
split.
fold (dist_Rp xs X0).
have H1:= not_series.
rewrite /conv_Rp in Hconv.
replace X0 with (Xn 0%nat).
apply seq_lim_in_dom.
apply Rmult_le_pos; [fourier|apply b0_ge_0].
by rewrite /conv_Rp.
exact H1.
by rewrite //=.
assert (Hxn: forall n, 
f (Xn n) = mult_mv (f' (Xn n))  ((Xn n) -^ (Xn (S n)))).
have H1: forall n,  mult_mv ((f' (Xn n))^-1m) (f (Xn n)) = ((Xn n) -^ (Xn (S n))).
intros n.
simpl.
rewrite /dif_v  -ext_eq => i; rewrite !vxK; by simp_R.
move=> n; rewrite -(H1 n) mult_mv_assoc.
destruct (about_seq n) as [_ [Hf' _]].
have Hf'' : \det (f' (Xn n)) !=0.
case X: (\det (f' (Xn n)) == 0); rewrite //=.
move/eqP: X =>X.
have H3 := Hf' X.
done.
by rewrite (mulVmx Hf'') unitM_el_neu.
have Hcv: conv_Rp (fun i:nat=> mult_mv (f' (Xn i)) ((Xn i) -^ (Xn (S i)))) (vect0 p).
apply lim_prod_minus with xs;[|done].

move=> a i.
apply  rel_cont_R_Rp => j.
apply  cont_eq_fun with ((fun x : Base (Rp_met p ) => der_p p  f j x (pr x j) i)).
move=> x.
by rewrite /f' /Jac !vxK !mxK.
by apply rel_cont_Rp_R_Rp.
have Hcv1: conv_Rp (fun i:nat =>f (Xn i)) (vect0 p).
rewrite /conv_Rp in Hcv |-*.
intros eps0 He0.
elim (Hcv eps0 He0); intros N Habs.
exists N.
intros n Hneq.
rewrite (Hxn n).
exact (Habs n Hneq).
have Hcv2: conv_Rp (fun i:nat => f (Xn i)) (f xs).
apply Heine1_Rp ;  done.
apply seq_lim_Rp_uniq with (fun i : nat => f (Xn i)).
exact Hp.
done.
done.
Qed.

End Kantorovitch.

(*Remarks:
for the proofs i have 3 admitted lemmas
 - 2 for matrices
 - 1 a consequence of taylor's formula
and 1 redundant hypothesis
 - the continuity of a diferentiable function
*)                                                                

