Require Import ssreflect ssrbool funs eqtype ssrnat seq fintype.
Require Import tuple paths connect div .
Require Import groups group_perm zp ssralg ssrbig matrix.
Require Import structRp min_max.
Require Import Reals.
Require Import Fourier.
Require Import Ranalysis.
Require Import Classical.
Require Import ClassicalChoice.
Require Import vectors seqRp.
Require Import Max.



Ltac simp_R:=
match goal with 
| |- (_ = _ )=> rewrite /(@Ring.opp ring_add_grp ) /=
                        /(@Ring.add ring_add_grp ) /= 
                        /(@Ring.mul ring_base ) /=
                       /0 /= /1 /=; ring_simplify
| _ =>  rewrite /(@Ring.opp ring_add_grp ) /=
                        /(@Ring.add ring_add_grp ) /= 
                        /(@Ring.mul ring_base ) /=
                       /0 /= /1 /=
end.


Section mat0.

Delimit Scope ring_scope with Ri.
Open Local Scope R_scope.

Variable p: pos_nat.
Lemma Hp: (O<p)%nat.
Proof.
have Hp1:= pos_natP p.
move/leP:Hp1=>Hp1.
omega.
Qed.



Notation "'MR_' p" := (matrix real_field p p)(at level 10).

Structure can_norm : Type := MNorm {
  anorm : MR_(p) -> R;
  pos_norm : forall A, R0 <= anorm A;
  hom_pos_norm : forall A r, anorm (r *s A) = Rabs r * anorm A;
  trin_ineq_norm : forall A B, anorm (A +m B) <= anorm A + anorm B;
  prod_ineq_norm: forall A B, anorm (A *m B) <= anorm A * anorm B;
  canonical_norm: forall (A:MR_(p)) i j, Rabs (A i j) <= anorm A;
  canonical_norm2: forall (A B:MR_(p)), (forall i j, Rabs (A i j) <= Rabs (B i j)) 
                                         ->anorm A <= anorm B
}.

Coercion Local anorm : can_norm  >-> Funclass.


Variable mynorm: can_norm.

Definition Rabs_M A:MR_(p) := \matrix_(i,j) Rabs (A i j).

Definition one_M: MR_(p) := \matrix_(i, j) R1.

Lemma mynorm0:
mynorm 0 = R0.
Proof.
have H1: mynorm 0 = mynorm (2 *s 0).
by rewrite scalemx0.
rewrite hom_pos_norm in H1.
case (classic (mynorm 0 = 0)) => H2;[done|].
have Hr: Rabs 2 = 1.
rewrite Rmult_comm -{1}(Rmult_1_r  (mynorm 0)) in H1;
by apply Rmult_eq_reg_l with (mynorm 0).
have Hr2: Rabs 2 <> R1.
rewrite Rabs_right; [apply Rgt_not_eq|]; fourier.
by case Hr2.
Qed.

Lemma RabsM_ineq:
forall (A:MR_(p)) eps, (forall i j, Rabs (A i j) < eps) ->
  mynorm A <= eps * mynorm one_M.
Proof.
move=> A e H.
have H1: forall i j, Rabs (A i j) < (e *s one_M) i j.
move=> i j; rewrite !mxK; simp_R; rewrite Rmult_1_r; apply H.
have Hepos: forall i j, Rabs ((e *s one_M) i j) = (e *s one_M) i j.
move=> i j; rewrite !mxK; simp_R; rewrite Rabs_right; first by done.
apply Rle_ge; apply Rle_trans with (Rabs (A i j)); [apply Rabs_pos| left; apply H] .
apply Rle_trans with (mynorm (e *s one_M)).
apply canonical_norm2.
move=> i j; rewrite Hepos; left; apply H1.
rewrite -{2}(Rabs_right e).
right; apply hom_pos_norm.
apply Rle_ge; apply Rle_trans with (Rabs (A (I0 p Hp) (I0 p Hp))); [apply Rabs_pos| left; apply H] .
Qed.




Definition dif_m  (A B:MR_(p)) :=
\matrix_(i, j) (A i j - B i j).

Lemma dif_mult_sc_1:
forall A B, dif_m A B  = A +m (-1 *s B).
Proof.
move=> A B; rewrite /dif_m /scalemx /addmx.
by apply/matrixP => i j; rewrite !mxK; simp_R.
Qed.

Notation "A -m B" := (dif_m A B) (at level 50, format "A  -m  B").
Notation "'\max_' ( i < p | P ) F" := (\big[maxn/O%N]_(i < p | P) F): dnat_scope.
 Notation "'\max_' ( i < p ) F" := (\big[maxn/O%N]_(i < p) F):dnat_scope.

Definition limit_m (An: nat -> MR_(p)) (La:MR_(p)):=
forall i j, Un_cv (fun n => An n i j) (La i j).

Lemma lem_cv1:
forall (Ak: nat -> MR_(p)) (A:MR_(p)), limit_m Ak A <-> Un_cv (fun k => mynorm (A -m (Ak k))) R0.
Proof.
move=> Ak A; split; move=> H.
red in H; red in H.
have H1: forall eps, eps>0->forall i j,  exists N : nat,
      forall n : nat, (n >= N)%nat -> R_dist (Ak n i j) (A i j) < eps.
by move=>*; apply H.
have H2: forall eps, eps>0-> forall i , exists c2:I_(p)->nat, 
forall j,   forall n : nat, (n >= (c2 j))%nat -> R_dist (Ak n i j) (A i j) < eps.
move=> eps He i; apply choice with (A:= I_(p)) (B:= nat) 
(R:=fun j N  =>  forall n : nat, (n >= N)%nat -> R_dist (Ak n i j) (A i j) < eps).
by move=> *; apply H1.
have H3: forall eps, eps>0-> exists c1: I_(p)->I_(p)->nat,
forall i j n,  (n >= (c1 i j))%nat -> R_dist (Ak n i j) (A i j) < eps.
move=> eps He .

apply (choice _ (H2 eps He)).
move=> eps He /=.
elim (Req_dec (mynorm one_M) R0) => Hdec.
elim (H3 eps He)=> ch Hch.
exists (\max_(i<p) \max_(j<p) ch i j)=> n Hn.
 apply Rle_lt_trans with R0.
rewrite /R_dist Rminus_0_r Rabs_right.
replace R0 with (eps * mynorm one_M).
apply RabsM_ineq => i j; rewrite !mxK Rabs_minus_sym.
apply:Hch.
rewrite /ge; apply le_trans with  (\max_(i < p) \max_(j < p) ch i j)%nat.
apply le_trans with  (\max_(j < p) ch i j)%nat.
apply: max_greater.
apply: (max_greater p (fun i => \max_(j0 < p) ch i j0 ) i).
 omega.
rewrite Hdec; ring.
apply Rle_ge;  apply: pos_norm.
done.

have Hen: 1/2* eps / mynorm one_M > 0.
apply: Rmult_gt_0_compat; 
[apply Rmult_gt_0_compat; simp_R; fourier |apply: Rinv_0_lt_compat].
elim (pos_norm mynorm  one_M) => Hen2;[done|by elim Hdec].
elim (H3 (1/2*eps/(mynorm one_M)) Hen)=> ch Hch.
exists (\max_(i<p) \max_(j<p) ch i j)=> n Hn.
rewrite /R_dist Rminus_0_r Rabs_right;[|apply Rle_ge; apply:pos_norm].
apply Rle_lt_trans with ((1/2*eps/ mynorm one_M) * mynorm one_M).
apply RabsM_ineq => i j; rewrite !mxK Rabs_minus_sym.
apply:Hch.
rewrite /ge; apply le_trans with  (\max_(i < p) \max_(j < p) ch i j)%nat.
apply le_trans with  (\max_(j < p) ch i j)%nat.
apply: max_greater.
apply: (max_greater p (fun i => \max_(j0 < p) ch i j0 ) i).
omega.
rewrite {1}/Rdiv Rmult_assoc Rinv_l; last by done .
rewrite Rmult_1_r ; simp_R; fourier.

move=> i j eps He.
elim (H eps He) => N HN.
exists N => n Hn.
rewrite /R_dist in HN |-*.
apply Rle_lt_trans with (mynorm (A -m Ak n)).
rewrite Rabs_minus_sym.
have Heq: (A i j - Ak n i j) = ((A -m Ak n) i j);
[by rewrite !mxK|].
rewrite Heq.
apply canonical_norm.
have H1:= HN n Hn.
rewrite Rminus_0_r in H1.
rewrite -(Rabs_right (mynorm (A -m Ak n)));[done|apply Rle_ge; apply:pos_norm].
Qed.

Delimit Scope dnat_scope with Dn.

Lemma echiv_sums:
forall N f, sum_f_R0 f N = \sum_(i<S N) f i.
Proof.
move => N f .
induction N.
rewrite big_ord_recl //=.
have Heq:  \sum_(i < O%nat) f (bump O%nat i) = R0.
unlock reducebig.
have Hind: index_enum I_(O%nat) = seq0.
rewrite /index_enum /enum //=.
rewrite Hind //=.
rewrite Heq; by simp_R.
rewrite //=.
have Heq: \sum_(j <= N.+1) f j =
f (N.+1) + \sum_(j <= N) f j  .
have H:= @bigD1 _ _ _ _ (@ord_max (N.+1)) (fun _ => true) (fun j => f j).
rewrite H //=.
simp_R.
apply Rplus_eq_compat_l.
rewrite -(big_mkord (fun i0 =>i0 != ord_max) (fun j => f j)).
have Heq: ((N.+1) <= (N.+1).+1)%Dn.
apply /leP; omega.
rewrite -(big_mkord (fun _ =>true) (fun j => f j)).
rewrite (big_nat_widen (O%nat) ((N.+1)%Dn) 
(((N.+1).+1)%Dn) (fun _ =>true) (fun j => f j) Heq) .
rewrite (big_mkord (fun i0 =>i0 != ord_max) (fun j => f j)).
rewrite (big_mkord (fun i0 =>true&& (i0 < N.+1)%Dn) (fun j => f j)).
apply: eq_bigl => j.

rewrite andTb.
case C: ((j < N.+1)%Dn).
move/leP:C=>C.
have Heq1: j<>ord_max.
move => Hc.
rewrite /ord_max in Hc.
have Heq2: (j.+1 = (N.+1).+1)%Dn.
f_equal.
by rewrite Hc /nat_of_ord.
rewrite Heq2 in C.
have Hc2: (~ (  ((N.+1).+1 <= N.+1)%nat)).
by omega.
by elim Hc2.
move/eqP: Heq1=>Heq1.
by rewrite Heq1.
move/leP:C=>C.
have Heq2: (j<(N.+1).+1)%Dn.
apply ordinal_ltn.
move/leP:Heq2=>Heq2.
have Heq3: ((j:nat) = (N.+1))%Dn.
omega.
rewrite Heq3 /ord_max /nat_of_ord.
apply negbET.
apply/ negP.
move/eqP=>P.
by elim P.
by rewrite Heq  IHN Rplus_comm.
Qed.

Definition cv_mat_ser (Ak:nat->MR_(p)) (A:MR_(p)) := 
forall i j, Un_cv (fun N => sum_f_R0 (fun n => Ak n i j) N) (A i j).

Definition cv_mat_ser2 (Ak:nat->MR_(p)) (A:MR_(p)) :=
limit_m (fun N => \sum_(i<N.+1) Ak i) A.

Lemma el_eq_seq_eq:
forall xn yn x , (forall n, xn n = yn n )-> Un_cv xn x-> Un_cv yn x.
Proof.
by move=> xn yn x Heq Hcv eps He; elim (Hcv eps He) => N HN;
exists N=> n Hn; rewrite -Heq; apply HN.
Qed.

Lemma el_eq_mat_seq_cv:
forall xn yn x, (forall n, xn n = yn n) -> limit_m xn x -> limit_m yn x.
Proof.
move=> xn yn x Hxy Hx i j.
apply el_eq_seq_eq with (fun n => xn n i j).
by move=> n; rewrite Hxy.
apply Hx.
Qed.

Lemma echiv_mat_ser:
forall Ak A, cv_mat_ser Ak A <-> cv_mat_ser2 Ak A.
Proof.
move=> Ak A; split=> H i j; have H1:= H i j.
have H2: forall N,  sum_f_R0 (fun k : nat => Ak k i j) N = 
(\sum_(i0 <= N) Ak i0) i j.
move=> N; rewrite echiv_sums.
unlock reducebig.
elim (index_enum I_(N.+1)) => //=.
by rewrite mxK.
by move => k s Ih; rewrite !mxK Ih.
apply (el_eq_seq_eq _ _ _ H2 H1).
have H2: forall N,  (\sum_(i0 <= N) Ak i0) i j
=sum_f_R0 (fun k : nat => Ak k i j) N .
move=> N; rewrite echiv_sums.
unlock reducebig.
elim (index_enum I_(N.+1)) => //=.
by rewrite mxK.
by move => k s Ih; rewrite !mxK Ih.
apply (el_eq_seq_eq _ _ _ H2 H1).
Qed.

Lemma th1_246:
forall Ak,  (exists S, cv_mat_ser2 Ak S) -> limit_m Ak \0.
Proof.
move=> Ak [S Hcv].
rewrite -echiv_mat_ser in Hcv.
move=> i j eps He.
have Hc2:= Hcv i j.
have Hcau: Cauchy_crit (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N).
by apply CV_Cauchy; exists (S i j).
elim (Hcau eps He) => N HN.
exists (N.+1)%nat => n Hn.
have H1: (n.-1>=N)%nat; [omega|].
have H2: ((n.-1).+1 >= N)%nat;[omega|].
have HN2:= HN _ _ H1 H2.
rewrite /R_dist //=  Rabs_minus_sym Rplus_comm 
/Rminus Rplus_assoc Rplus_opp_r Rplus_0_r in HN2.
rewrite /R_dist mxK Rminus_0_r.
replace n with ((n.-1).+1);[done|omega].
Qed.

Lemma sum_S_eq:
forall (f:nat -> R) n,
\sum_(i<=n.+1) f i = f (n.+1) + \sum_(i<=n) f i.
Proof.
move=> f N.
have H:= @bigD1 _ _ _ _ (@ord_max (N.+1)) (fun _ => true) (fun j => f j).
rewrite H //=.
simp_R.
apply Rplus_eq_compat_l.
rewrite -(big_mkord (fun i0 =>i0 != ord_max) (fun j => f j)).
have Heq: ((N.+1) <= (N.+1).+1)%Dn.
apply /leP; omega.
rewrite -(big_mkord (fun _ =>true) (fun j => f j)).
rewrite (big_nat_widen (O%nat) ((N.+1)%Dn) 
(((N.+1).+1)%Dn) (fun _ =>true) (fun j => f j) Heq) .
rewrite (big_mkord (fun i0 =>i0 != ord_max) (fun j => f j)).
rewrite (big_mkord (fun i0 =>true&& (i0 < N.+1)%Dn) (fun j => f j)).
apply: eq_bigl => j.

rewrite andTb.
case C: ((j < N.+1)%Dn).
move/leP:C=>C.
have Heq1: j<>ord_max.
move => Hc.
rewrite /ord_max in Hc.
have Heq2: (j.+1 = (N.+1).+1)%Dn.
f_equal.
by rewrite Hc /nat_of_ord.
rewrite Heq2 in C.
have Hc2: (~ (  ((N.+1).+1 <= N.+1)%nat)).
by omega.
by elim Hc2.
move/eqP: Heq1=>Heq1.
by rewrite Heq1.
move/leP:C=>C.
have Heq2: (j<(N.+1).+1)%Dn.
apply ordinal_ltn.
move/leP:Heq2=>Heq2.
have Heq3: ((j:nat) = (N.+1))%Dn.
omega.
rewrite Heq3 /ord_max /nat_of_ord.
apply negbET.
apply/ negP.
move/eqP=>P.
by elim P.
Qed.

(*
Lemma sum_S_eq2:
forall (f:nat -> R) m n,
\sum_(m<=i<n.+1) f i = f (n) + \sum_(m<=i<n) f i.
Proof.
move=> f m N.
have H:= @bigD1 _ _ _ _ (N:nat_eqType) (fun _ => true) (fun j => f j).
rewrite H.
simp_R.
apply Rplus_eq_compat_l.
rewrite -(big_mkord (fun i0 =>i0 != ord_max) (fun j => f j)).
have Heq: ((N.+1) <= (N.+1).+1)%Dn.
apply /leP; omega.
rewrite -(big_mkord (fun _ =>true) (fun j => f j)).
rewrite (big_nat_widen (O%nat) ((N.+1)%Dn) 
(((N.+1).+1)%Dn) (fun _ =>true) (fun j => f j) Heq) .
rewrite (big_mkord (fun i0 =>i0 != ord_max) (fun j => f j)).
rewrite (big_mkord (fun i0 =>true&& (i0 < N.+1)%Dn) (fun j => f j)).
apply: eq_bigl => j.

rewrite andTb.
case C: ((j < N.+1)%Dn).
move/leP:C=>C.
have Heq1: j<>ord_max.
move => Hc.
rewrite /ord_max in Hc.
have Heq2: (j.+1 = (N.+1).+1)%Dn.
f_equal.
by rewrite Hc /nat_of_ord.
rewrite Heq2 in C.
have Hc2: (~ (  ((N.+1).+1 <= N.+1)%nat)).
by omega.
by elim Hc2.
move/eqP: Heq1=>Heq1.
by rewrite Heq1.
move/leP:C=>C.
have Heq2: (j<(N.+1).+1)%Dn.
apply ordinal_ltn.
move/leP:Heq2=>Heq2.
have Heq3: ((j:nat) = (N.+1))%Dn.
omega.
rewrite Heq3 /ord_max /nat_of_ord.
apply negbET.
apply/ negP.
move/eqP=>P.
by elim P.
Qed.

*)
(*this might already exist in the library*)
Lemma sum_dif:
forall (f:nat ->R) m n, (m < n)%nat ->
 \sum_(i<=n) f i - \sum_(i<=m) f i  = \sum_(m.+1 <= i < n.+1) f i.
Proof.
move=> f m n.
induction n => H //=.
have H1: m=O;[omega|].
rewrite H1.
rewrite big_add1.
have H2: 1.-1 = O;[omega|].
rewrite H2 big_mkord.
rewrite /Rminus Rplus_opp_r.
have Heq:  \sum_(i < O%nat) f i.+1 = R0.
unlock reducebig.
have Hind: index_enum I_(O%nat) = seq0.
rewrite /index_enum /enum //=.
rewrite Hind //=.
rewrite Heq; by simp_R.
rewrite //=.
have H0 : (m <= n)%nat; [by omega|].
elim (le_lt_eq_dec _ _ H0) => H1.
rewrite sum_S_eq.
transitivity ( \sum_(i <= n) f i - \sum_(i <= m) f i + f (n.+1));
[rewrite //=; ring|].
rewrite (IHn H1).
replace (m.+1) with (O+(m.+1))%nat;[|omega].
have H2:= (@big_addn real_ring _ _ (O%nat) (n.+1)%nat (m.+1) (fun _ => true) f).
rewrite H2.
have H3:= (@big_addn real_ring _ _ (O%nat) ((n.+1).+1) (m.+1) (fun _ => true) f).
rewrite H3.
rewrite !big_mkord.
replace (n.+1 - m.+1)%Dn with ((n - m.+1).+1)%Dn.
replace ( (n.+1).+1 - m.+1)%Dn with ((((n - m.+1).+1).+1)%nat).
have H4:= sum_S_eq (fun i => f (i + m.+1)%Dn) ((n - m.+1)).
rewrite H4.
replace (((n - m.+1).+1 + m.+1)%Dn) with (n.+1).
rewrite //=; ring.
rewrite -subnI -addnI;omega.
rewrite -subnI ;omega.
rewrite -subnI ;omega.
rewrite H1.
rewrite sum_S_eq /Rminus Rplus_assoc Rplus_opp_r Rplus_0_r.
have H2:= (@big_addn real_ring _ _ (O%nat) ((n.+1).+1) ((n.+1)) (fun _ => true) f).
pattern (n.+1) at 2; replace (n.+1) with (O%nat+n.+1)%nat;[|omega].
rewrite H2.
rewrite big_mkord.
replace ((n.+1).+1 - n.+1)%Dn with (S O)%nat.
rewrite big_ord_recl //=.
have Heq:  \sum_(i < O%nat) f (bump O%nat i + n.+1)%Dn = R0.
unlock reducebig.
have Hind: index_enum I_(O%nat) = seq0.
rewrite /index_enum /enum //=.
rewrite Hind //=.
rewrite Heq;  simp_R; f_equal.
rewrite -subnI ;omega.
Qed.

Lemma abs_sum_le_f0:
forall (f:nat ->R) N, Rabs (\sum_(i<N) (f i)) <= \sum_(i<N) (Rabs (f i)). 
Proof.
move => v N; apply: (big_op_rel Rle Rabs);
[right; apply Rabs_R0
|exact Rle_trans
|by move=>*; simp_R; apply Rplus_le_compat_l
|move=>*; apply Rabs_triang].
Qed.


Lemma abs_cv_cv:
forall xn xs, Un_cv (fun n => sum_f_R0 (fun k => Rabs (xn k)) n) xs ->
exists l, Un_cv (fun n=> sum_f_R0 xn n) l.
Proof.
move=> xn xs Hcva.
have H1: Cauchy_crit (fun n : nat => sum_f_R0 (fun k : nat => Rabs (xn k)) n).
apply CV_Cauchy; by exists xs.
have H2: sigT (fun l => Un_cv (fun n : nat => sum_f_R0 xn n) l).
apply R_complete => eps He.
elim (H1 eps He) => N HN.
exists N => n m Hn Hm; have H2:= HN n m Hn Hm.
rewrite /R_dist in H2 |-*.
rewrite !echiv_sums in H2 |-*.
elim (le_lt_dec m n) => Hdec.
elim (le_lt_eq_dec _ _ Hdec) => Hd2.
rewrite (sum_dif _ _ _ Hd2).
have H3:= (@big_addn real_ring _ _ (O%nat) (n.+1)%nat (m.+1) (fun _ => true) xn).
rewrite H3 big_mkord.
apply Rle_lt_trans with (\sum_(i < n.+1 - m.+1) Rabs (xn (i + m.+1)%Dn)).
apply (abs_sum_le_f0 (fun i =>xn (i + m.+1)%Dn) (n.+1 - m.+1)) .
have H4:= (@big_addn real_ring _ _ (O%nat) (n.+1)%nat (m.+1) 
   (fun _ => true) (fun i => Rabs (xn i))).
rewrite (sum_dif ( fun i => Rabs (xn i)) _ _ Hd2) H4 big_mkord in H2.
apply Rle_lt_trans with (Rabs (\sum_(i < n.+1 - m.+1) Rabs (xn (i + m.+1)%Dn)) ).
 apply RRle_abs .
done.
by rewrite Hd2 /Rminus Rplus_opp_r Rabs_R0.
rewrite Rabs_minus_sym .
rewrite (sum_dif _ _ _ Hdec).
have H3:= (@big_addn real_ring _ _ (O%nat) (m.+1)%nat (n.+1) (fun _ => true) xn).
rewrite H3 big_mkord.
apply Rle_lt_trans with (\sum_(i < m.+1 - n.+1) Rabs (xn (i + n.+1)%Dn)).
apply (abs_sum_le_f0 (fun i =>xn (i + n.+1)%Dn) (m.+1 - n.+1)) .
have H4:= (@big_addn real_ring _ _ (O%nat) (m.+1)%nat (n.+1) 
   (fun _ => true) (fun i => Rabs (xn i))).
rewrite Rabs_minus_sym (sum_dif ( fun i => Rabs (xn i)) _ _ Hdec) H4 big_mkord in H2.
apply Rle_lt_trans with (Rabs (\sum_(i < m.+1 - n.+1) Rabs (xn (i + n.+1)%Dn)) ).
 apply RRle_abs .
done.
by elim H2 => l Hl; exists l.
Qed.


Lemma th2_246:
forall (Ak: nat -> MR_(p)), (exists S,  cv_mat_ser (fun n=> Rabs_M (Ak n)) S) -> exists S',
cv_mat_ser Ak S'.
Proof.
move=> Ak  [S Hcv].
 red in Hcv.
rewrite /cv_mat_ser .
have H1:  forall i j, exists l,  Un_cv 
  (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N) l.
move=> i j.
apply abs_cv_cv with (S i j).
apply el_eq_seq_eq with (fun N : nat => sum_f_R0 (fun n : nat => Rabs_M (Ak n) i j) N).
rewrite /Rabs_M.
induction n => //=; [by rewrite !mxK|by rewrite IHn !mxK].
apply Hcv.
have Hc: forall i, exists c1:I_(p)->R, forall j,
 Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N) (c1 j).
by move=> i; apply choice with (R:= fun j l=> 
 Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N) l) .
have Hc2: exists c2:I_(p)->I_(p)->R, forall i j, 
Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Ak n i j) N) (c2 i j).
apply (choice _ (Hc)).
elim Hc2 => ch Hch.
by exists (\matrix_(i, j) ch i j) => i j; rewrite !mxK.
Qed.

Lemma thm2_246:
forall (Ak: nat -> MR_(p)) , (exists S, cv_mat_ser2 (fun n=> Rabs_M (Ak n)) S) -> exists S',
cv_mat_ser2 Ak S'.
Proof.
move=> Ak  [S Hs].
have Hs1: exists S : matrix real_field p p,
       cv_mat_ser (fun n : nat => Rabs_M (Ak n)) S.
by exists S; rewrite -echiv_mat_ser in Hs.
elim (th2_246 _  Hs1) => S' HS'.
by exists S'; rewrite -echiv_mat_ser.
Qed.

Lemma th3_247:
forall Ak S, Un_cv (fun N:nat => sum_f_R0 (fun n => mynorm (Ak n))  N) S -> 
exists S', cv_mat_ser (fun n=> Rabs_M (Ak n)) S'.
Proof.
move=> Ak S Hcv.
rewrite /cv_mat_ser.
have H: forall i j, sigT (fun l =>  
 Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Rabs_M (Ak n) i j) N) l).
move=> i j.
apply Rseries_CV_comp with 
 ((fun n : nat => mynorm (Ak n) ) ).
move=> n; rewrite  !mxK; split; [apply Rabs_pos|apply canonical_norm].
by exists S.
have H1: forall i j, exists l,
 Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Rabs_M (Ak n) i j) N) l.
by move=> i j; elim (H i j) => l  Hl; exists l.
have H2: forall i, exists S1,
  forall j : I_(p),
  Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Rabs_M (Ak n) i j) N)
    (S1  j).
move=> i; apply (choice _ (H1 i)).
have H3: exists S1,
  forall i j : I_(p),
  Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Rabs_M (Ak n) i j) N)
    (S1 i j).
apply (choice _ H2).
by elim H3 => S1  Hs1;  exists (\matrix_(i, j) S1 i j)=> i j; rewrite !mxK.
Qed.

Delimit Scope matrix_scope with mx.

Hypothesis Hyp_norm1:
mynorm 1 = 1. 

Lemma matr_prod_ineq:
forall X k, mynorm (X^+k) <= (mynorm X)^+k.
Proof.
move=> X k; elim k => //=.
by right.
move=> n IHn.
have H1:= prod_ineq_norm mynorm X (X^+n).
rewrite /(@Ring.mul) //=.
apply Rle_trans with (mynorm X * mynorm (X^+n)).
done.
apply Rmult_le_compat_l.
apply pos_norm.
done.
Qed.



Lemma th4_247:
forall X, mynorm X < R1 ->  exists S,  cv_mat_ser
          (fun n =>  X^+n) S.
Proof.
move=> X Hx.
apply th2_246.
rewrite /cv_mat_ser.
have H1: forall i j, sigT (fun S =>
  Un_cv (fun N : nat => sum_f_R0 (fun n : nat =>Rabs_M  (X ^+n) i j) N) S ).
move=> i j.
apply Rseries_CV_comp with (fun n => (mynorm X)^+n).
move=> n; rewrite !mxK; split; [apply Rabs_pos|].
apply Rle_trans with (mynorm (X^+n)).
apply canonical_norm.
apply matr_prod_ineq.
have H3: Rabs (mynorm X) < 1.
rewrite Rabs_right;[done|apply Rle_ge; apply: pos_norm].

have H2:= GP_infinite (mynorm X) H3 .
exists (/(1 - mynorm X)).
move=> eps He; case (H2 eps He)=> N Hn0; exists N => n Hn.
have H4:= Hn0 n Hn.

rewrite /R_dist in H4 |-*.
apply Rle_lt_trans with 
 (Rabs (sum_f_R0 (fun n : nat => 1 * mynorm X ^ n) n - / (1 - mynorm X)) ).
right.
congr (Rabs _).
congr ( _ - _).
elim n => //=.
by simp_R.
move=> n0 Ihn; rewrite Ihn; simp_R.
congr ( _ + _ ); congr ( _ * _).
elim n0 => //=; move=> n1 Ihn1; simp_R; by congr ( _ * _ ).
done.
have Ha: forall i j : I_(p),
     exists S, 
     (Un_cv (fun N : nat => sum_f_R0 (fun n : nat => Rabs_M (X ^+n) i j) N) S).
by move=> i j; case (H1 i j) => S Hs; exists S.
have Hz:  forall i : I_(p),
     exists S, forall j, 
     (Un_cv (fun N : nat => 
              sum_f_R0 (fun n : nat => Rabs_M (X ^+n) i j) N) (S j)).
move=> i; case (choice _ (Ha i)) => ch1 Hch1; by exists ch1.
case (choice _ Hz) => ch Hch.
by exists (\matrix_(i, j) ch i j) => i j; rewrite !mxK.
Qed.

Lemma sum_prod_id:
forall (X:MR_(p)) k , ((1 - X)*(\sum_(0<=i < k.+1) (X^+i))  = 1 - X^+(k.+1))%Ri.
Proof.
move=> X k.
replace (1 - X)%Ri with (- (-1 + X))%Ri;[|by rewrite Ring.oppr_add Ring.opprK].
rewrite Ring.mulNr  Ring.mulr_addl  Ring.mulN1r Ring.oppr_add Ring.opprK.
rewrite big_distrr  //= -sum_split_sub //=.
induction k => //=.
rewrite big_mkord.
rewrite big_ord_recl //= .
have Heq:  (\sum_(i < O%nat) (X * X ^+i - X *(X* X ^+i )))%Ri = 0.
unlock reducebig.
have Hind: index_enum I_(O%nat) = seq0.
rewrite /index_enum /enum //=.
rewrite Hind //=.
by rewrite Heq Ring.addr0 . 
rewrite //=.
rewrite !big_mkord in IHk |-*. 
by rewrite big_ord_recr //= IHk Ring.addrA Ring.mulrA Ring.addrC -Ring.addrA
Ring.addNr Ring.addr0 Ring.addrC.
Qed.


Lemma mat_to_zero:
forall X, mynorm X < 1 -> limit_m (fun n => X^+n) \0.
Proof.
move=> X Hx.
apply th1_246.
have H1: exists S0 : matrix real_field p p, cv_mat_ser (fun n : nat => X ^+n) S0.
by apply th4_247.
by case H1 => S Hs; exists S;
apply/echiv_mat_ser.
Qed.

Lemma abs_sum_le_f1:
forall (f:I_(p) ->R) , Rabs (\sum_(i) (f i)) <= \sum_(i) (Rabs (f i)). 
Proof.
move => v.
unlock reducebig.
move : (I_not_void p Hp );rewrite /index_enum.
elim : (enum I_(p)) (uniq_enum I_(p)) => //=.
move => x s IH IHun _ .
apply Rle_trans with (Rabs (v x) + Rabs (foldr (fun (i : ordinal p) (x0 : R) => v i + x0) 0 s) ).
apply Rabs_triang.
apply Rplus_le_compat_l.
move/andP : IHun => [In Iu].
case C:(seq0!=s).
apply (IH Iu C).
move/eqP: C=>C.
rewrite -C //= Rabs_R0; apply Rle_refl.
Qed.

Lemma limit_m_mul:
forall A R S, limit_m A S -> limit_m (fun n => R * (A n))%Ri (R * S)%Ri.
Proof.
move=> A R S Hl i j eps He.
red in Hl; red in Hl.
have Hl1: forall eps, eps > R0 -> forall (i j : I_(p)),
     exists N : nat,
       forall n : nat, (n >= N)%nat -> R_dist (A n i j) (S i j) < eps.
by move=>*; apply Hl.
have H0 : R0 <= \big[Rmax/R0]_(j) Rabs (R i j) .
case (max_exists p Hp (fun j =>  Rabs (R i j)))=> i1;
 [apply Rabs_pos| 
move=> H; rewrite H; apply Rabs_pos].
case H0 => H.
have H01: 0 < \big[Rmax/R0]_(i) \big[Rmax/0]_(j) Rabs (R i j).
apply Rlt_le_trans with (\big[Rmax/0]_(j) Rabs (R i j)); [done|].
apply (max_max p (fun i0 =>\big[Rmax/0]_(j0) Rabs (R i0 j0))) => i0.
case (max_exists p Hp (fun j0=>  Rabs (R i0 j0)))=> i1;
 [apply Rabs_pos| 
move=> H02; rewrite H02; apply Rabs_pos].

have H1: (eps / (\big[Rmax/0]_(i) \big[Rmax/R0]_(i0) Rabs (R i i0))) /INR p > R0.
rewrite /Rdiv; apply Rmult_gt_0_compat.
 apply Rmult_gt_0_compat; [done|by apply: Rinv_0_lt_compat] .
apply: Rinv_0_lt_compat; apply: lt_INR_0; apply Hp.
have Hl2:= (Hl1 ((eps/(\big[Rmax/R0]_(i) \big[Rmax/R0]_(i0) (Rabs (R i i0))))/INR p) H1).
have Hc1:  forall i : I_(p), exists c1: I_(p) -> nat,
        forall j n,
        (n >= (c1 j))%nat ->
        R_dist (A n i j) (S i j) <
        eps / \big[Rmax/0]_(i0) \big[Rmax/0]_(i1) Rabs (R i0 i1) / INR p.
move=> i0; apply (choice _ (Hl2 i0)).
case (choice _ Hc1) => ch Hch.
exists ( \max_(i<p)  \max_(j<p) (ch i j)) => n Hn.
rewrite !mxK /R_dist.
apply Rle_lt_trans with (Rabs (\sum_(j0) (R i j0 * A n j0 j)%Ri - \sum_(j0) (R i j0 * S j0 j)%Ri)%Ri).
right; f_equal.
rewrite -sum_split_sub.
apply Rle_lt_trans with (Rabs (\sum_(i0) (R i i0 * (A n i0 j -  S i0 j)%Ri))).
by right; f_equal; apply eq_bigr => i0 _; simp_R; ring_simplify.
apply Rle_lt_trans with  (\sum_(i0) Rabs (R i i0 * (A n i0 j - S i0 j)%Ri)).
rewrite //=; apply abs_sum_le_f1 with (f:= (fun i0 => R i i0 * (A n i0 j - S i0 j))) .
apply Rle_lt_trans with (\sum_(i0) (\big[Rmax/0]_(i0) Rabs (R i i0) * Rabs (A n i0 j - S i0 j)%Ri)).
apply (sum_le_f_le) => i0.
rewrite Rabs_mult.
apply Rmult_le_compat.
apply Rabs_pos .
apply Rabs_pos.
apply (max_max p (fun j=> Rabs (R i j)))=> i1; apply Rabs_pos.
by right.
rewrite -big_distrr //= Ring.mulrC.

replace eps with (  (eps / \big[Rmax/0]_(i0) Rabs (R i i0)) * \big[Rmax/0]_(i0 < p) Rabs (R i i0) ).
apply Rmult_lt_compat_r .
done.
apply: (ineq_sum_el Hp).
left; apply: Rmult_gt_0_compat; [done|by apply: Rinv_0_lt_compat] .
move=> x.
apply Rlt_le_trans with (eps / \big[Rmax/0]_(i0) \big[Rmax/0]_(i1) Rabs (R i0 i1) / INR p).
rewrite /R_dist in Hch.
 apply: Hch.
apply: (le_trans _  (\max_(i < p) \max_(j < p) ch i j)%nat _);[|apply Hn].
apply le_trans with (\max_(j0 < p) ch x j0 ).
apply (max_greater p (fun j0 => ch x j0)).
apply (max_greater p (fun i0 =>\max_(j0 < p) ch i0 j0) ) .
apply: Rmult_le_compat_r;[left; apply: Rinv_0_lt_compat; apply lt_INR_0; apply Hp|].
apply: Rmult_le_compat_l; [by left|].
apply Rle_Rinv; try done.
apply (max_max p (fun i0 => \big[Rmax/0]_(i1) Rabs (R i0 i1))) => i0.
case (max_exists p Hp (fun j =>  Rabs (R i0 j)))=> i1;
 [apply Rabs_pos| 
move=> Hx; rewrite Hx; apply Rabs_pos].
rewrite Rmult_assoc Rinv_l; [ring|].
by apply Rgt_not_eq.
exists O => n Hn.
rewrite !mxK /R_dist.
apply Rle_lt_trans with (Rabs (\sum_(j0) (R i j0 * A n j0 j)%Ri - \sum_(j0) (R i j0 * S j0 j)%Ri)%Ri).
right; f_equal.
rewrite -sum_split_sub.
apply Rle_lt_trans with (Rabs (\sum_(i0) (R i i0 * (A n i0 j -  S i0 j)%Ri))).
by right; f_equal; apply eq_bigr => i0 _; simp_R; ring_simplify.
apply Rle_lt_trans with  (\sum_(i0) Rabs (R i i0 * (A n i0 j - S i0 j)%Ri)).
rewrite //=; apply abs_sum_le_f1 with (f:= (fun i0 => R i i0 * (A n i0 j - S i0 j))) .
apply Rle_lt_trans with (\sum_(i0) (\big[Rmax/0]_(i0) Rabs (R i i0) * Rabs (A n i0 j - S i0 j)%Ri)).
apply (sum_le_f_le ) => i0.
rewrite Rabs_mult.
apply Rmult_le_compat.
apply Rabs_pos .
apply Rabs_pos.
apply (max_max p (fun j=> Rabs (R i j)))=> i1; apply Rabs_pos.
by right.
rewrite -big_distrr //= Ring.mulrC.
by rewrite -H Ring.mulr0.
Qed.

Lemma limit_m_add:
forall A B x y, limit_m A x -> limit_m B y -> limit_m (fun n => A n +m B n) (x +m y).
Proof.
move=>  a b x y Ha Hb i j.
rewrite !mxK.
apply el_eq_seq_eq with (fun n => a n i j + b n i j); [by move=> n; rewrite mxK; simp_R|].
apply CV_plus;[apply Ha|apply Hb].
Qed.

Lemma limit_m_opp_0:
forall a , limit_m a \0 -> limit_m (fun n=> (- a n)%Ri) \0.
Proof.
move=> a Hcv.
move=> i j eps He; case (Hcv  i j eps He) => N HN; exists N => n Hn.
have HN1 := HN n Hn.
rewrite !mxK /R_dist !Rminus_0_r in HN1 |- *.
by rewrite Ring.mulNr Rabs_Ropp Ring.mul1r.
Qed.

Lemma limit_m_Sn:
forall a x, limit_m a x -> limit_m (fun n => a (n.+1)) x.
Proof.
move=> a x Hcv i j eps He ; case (Hcv i j eps He) => N HN; 
exists (N.+1) => n Hn; apply HN; omega.
Qed.

Lemma limit_m_uniq:
forall a x y, limit_m a x -> limit_m a y -> x = y.
Proof.
move=> a x y  Hx Hy; apply/ matrixP => i j.
red in Hx.
by apply UL_sequence with (fun n => a n i j) .
Qed. 

Lemma th5_aj1:
forall X S, mynorm X < 1 -> cv_mat_ser2 (fun n => X^+n) S -> (\1 -m X) *m S = \1.
Proof.
move=> X S Hmn1 Hcv.
red in Hcv.
have Hcv2:=  (limit_m_mul _ (\1 - X)%Ri _ Hcv).
have Hcv3: limit_m (fun n : nat => ((\1 - X) * (\sum_(i <= n) X ^+i))%Ri)
         (\1)%Ri.
apply el_eq_mat_seq_cv with (fun n => ((\1 - X^+n.+1))%Ri).
by move=> n; rewrite  -sum_prod_id big_mkord.
replace (\1:matrix real_field p p) with ((\1 +m \0):matrix real_field p p).
apply: limit_m_add.
by move=> i j eps He; exists O=> n Hn; rewrite /R_dist !mxK Ring.addr0 /Rminus Rplus_opp_r
Rabs_R0.
apply limit_m_opp_0; apply limit_m_Sn ; rewrite //=.
by apply: mat_to_zero.
by apply/matrixP => i j; rewrite !mxK Ring.addr0.
apply limit_m_uniq with (fun n =>  ((\1 - X) * (\sum_(i <= n) X ^+i))%Ri); [|done] .
have Heq :  ((\1 -m X) *m S) =  ((\1 - X) * S)%Ri.
rewrite dif_mult_sc_1 /Ring.mul  //= .
by rewrite Heq .
Qed.



Lemma th5_248:
forall X, mynorm X < R1 -> \det (\1 -m X) <> 0.
Proof.
move=> X H.
case (th4_247 X H) => S HS.
rewrite echiv_mat_ser in HS.
have H1:= th5_aj1 X S H HS.
have H0: (\det (\1 -m X) * \det S)%Ri <> 0.
rewrite -(@determinantM _  p ((\1 -m X))  S) H1 determinant1; simp_R; auto with real. 
rewrite /Ring.mul //= in H0 .
by case (RIneq.Rmult_neq_0_reg (\det (\1 -m X)) (\det S) H0).
Qed.


Lemma cor1_th5:
forall X, mynorm (\1 -m X) < R1 -> \det X <> 0.
Proof.
move => X Hn.
have Heq: X = \1 -m (\1 -m X).
apply/matrixP => i j; rewrite !mxK {1}/Rminus.
by rewrite Ropp_minus_distr'  Rplus_comm Rplus_assoc Rplus_opp_l Rplus_0_r.
by rewrite Heq; apply th5_248.
Qed.

Notation "A ^-1m" := (invmx  A) (at level 10, format "A ^-1m") .

Lemma cor2_th5:
forall X, mynorm X < R1 ->  cv_mat_ser2 (fun n => X^+n) ((\1 -m X) ^-1m).
Proof.
move=> X H.
rewrite -echiv_mat_ser.
case (th4_247 X H) => S HS.
rewrite echiv_mat_ser in HS.
have Hm:= th5_aj1 X S H HS. 
have Heq2: (\1 -m X) *m ((\1 -m X)^-1m) = \1.
by apply: mulVmx ; apply /eqP; apply th5_248.
have Heq : S = ((\1 -m X)^-1m).
have Heq:  (\1 -m X) *m S =  (\1 -m X) *m (\1 -m X)^-1m.
by rewrite Hm Heq2.
have H1: \det (\1 -m X) != 0.
by apply/eqP; apply th5_248.
apply: (mulImx H1 Heq).
by rewrite -Heq echiv_mat_ser.
Qed.

Lemma tri_minus_norm:
forall a b, mynorm a - mynorm b <= mynorm (a - b)%Ri.
Proof.
move=> a b.
apply (Rplus_le_reg_l (mynorm b)).
ring_simplify.
pattern a at 1; replace a with (b +( a - b))%Ri.
apply: trin_ineq_norm.
rewrite Ring.addrCA Ring.addrA.
by rewrite Ring.addrK.
Qed.

Lemma eq_norm_min:
forall a b, mynorm (a - b)%Ri = mynorm (b - a)%Ri.
Proof.
move=> a b.
replace (a - b)%Ri with (-1 *s (b - a)%Ri).
by rewrite hom_pos_norm Rabs_Ropp Rabs_R1 Rmult_1_l.
by rewrite /Ring.add /Ring.opp //=; apply/matrixP=> i j; rewrite !mxK; simp_R.
Qed.

Lemma tri_min2:
forall a b, Rabs (mynorm a - mynorm b) <= mynorm (a - b)%Ri.
Proof.
move=> a b.
rewrite /Rabs; case (Rcase_abs (mynorm a - mynorm b) )=> H;
[|apply tri_minus_norm] .
rewrite Ropp_minus_distr eq_norm_min; apply tri_minus_norm.
Qed.

(*folowing 2 lemmas are from kantoro.v*)
Lemma lim_conv_low:
forall (un:nat->R) (a l:R),
(forall n, a<=un n)->
Un_cv un l ->
a<=l.
Proof.
intros.
cut (~(l<a)).
intros.
apply Rnot_gt_le; auto.
unfold not; intros.
unfold Un_cv in H0.
assert (a-l>0).
simp_R; fourier.
assert (Hcon: exists N, forall n, (n>=N)%nat ->R_dist (un n) l <a-l).
apply (H0 (a-l) H2).
elim Hcon; intros N Hlim.
assert (Hdist:R_dist (un N) l<a-l).
apply (Hlim N).
omega.
unfold R_dist in Hdist.
assert (Ha: a<=un N).
apply (H N).
rewrite Rabs_right in Hdist; [|fourier].
assert (un N<a); fourier.
Qed.

Lemma lim_seq_monot:
forall (xn yn:nat->R) (x y:R), 
(forall n, xn n<= yn n)->
(Un_cv xn x)->
(Un_cv yn y)->
x<=y.
intros.
assert (0<=y-x).
assert (forall n, 0<=yn n- xn n).
intros.
elim (H n);intros; simp_R; rewrite //=; fourier.
assert (Un_cv (fun i:nat => yn i - xn i) (y - x)).
apply CV_minus; assumption.
apply lim_conv_low with (fun  i:nat => yn i - xn i);
assumption.
rewrite /Ring.zero //= in H2; fourier.
Qed.

Lemma mat_norm_sum: forall (A : nat -> MR_(p)) n,
   mynorm (\sum_(i <= n) A i) <=  \sum_(i <= n) mynorm (A i).
Proof.
move=> A n.
apply: big_morph_rel.
by move=> x; right.
by move=> *; apply Rplus_le_compat.
by move=>b *; apply Rle_trans with b.
apply mynorm0.
by move=> a v; apply: trin_ineq_norm.
Qed.


Lemma norm_inv_ser0:
forall a, mynorm a < R1 -> mynorm (\1) = 1 ->
mynorm (( \1 -m  a)^-1m) <= 1/(1 - mynorm a).
Proof.
move=> X H H1.
have H2 := cor2_th5 X H.
red in H2; rewrite lem_cv1 in H2.
have H3: Un_cv (fun k : nat => mynorm ( \sum_(i <= k) X ^+i)) 
(mynorm ((\1 -m X)^-1m)) .
move=> eps He.
case (H2 eps He) => N HN; exists N => n Hn.
rewrite /R_dist in HN |-*.
have HN' := HN n Hn.
rewrite Rminus_0_r in HN'.
apply Rle_lt_trans with ( (mynorm ((\1 -m X)^-1m -m \sum_(i <= n) X ^+i)) ).
have Heq: mynorm ((\1 -m X)^-1m -m \sum_(i <= n) X ^+i) =
 mynorm ( \sum_(i <= n) X ^+i -m (\1 -m X)^-1m ).

have Hb:= eq_norm_min (\sum_(i <= n) X ^+i)  ((\1 -m X)^-1m).
rewrite !dif_mult_sc_1 in Hb |-*.
rewrite  /Ring.opp /Ring.one //= in Hb.
rewrite Heq.
have H7:= tri_min2 (\sum_(i <= n) X ^+i)  ((\1 -m X)^-1m).
rewrite !dif_mult_sc_1 in H7 |-*.
rewrite  /Ring.opp /Ring.one //= in H7.
rewrite -(Rabs_right (mynorm ((\1 -m X)^-1m -m \sum_(i <= n) X ^+i) ));
[done| apply Rle_ge ; apply pos_norm].
apply lim_seq_monot with ((fun k : nat => mynorm (\sum_(i <= k) X ^+i)))
 ((fun k : nat => (\sum_(i <= k) (mynorm X) ^+i))).
move=> n.
apply Rle_trans with (\sum_(i <= n) mynorm (X ^+i)).
apply: mat_norm_sum.
apply: big_rel.
by right.
by move=>*; apply Rplus_le_compat.
by move=> *; apply matr_prod_ineq.
(* add this statement as a new lemma: norm_sum_ineq
mynorm (\sum_(i <= n) X ^+i) <= \sum_(i <= n) mynorm X ^+i
proof using big_op_rel; give generalisation fu: R1->R2;
R1:= matrix R p p; R2:= R


unlock reducebig.
elim ((index_enum I_(n.+1))) => //=.
have Heq:  (mynorm 0) =  (mynorm (0 *s \1)).
f_equal; apply /matrixP=> i j; rewrite !mxK; simp_R; ring.
rewrite Heq hom_pos_norm Rabs_R0 Rmult_0_l; by right.
move=> N s IHs.
apply Rle_trans with ( mynorm
  (X ^+N) + mynorm (
   foldr (fun (i : ordinal n.+1) (x : matrix real_ring p p) => X ^+i + x) 0 s)%Ri ).
apply: trin_ineq_norm.
apply Rplus_le_compat.
apply matr_prod_ineq.
done.
*)
done.
have H4: Rabs (mynorm X) < 1.
rewrite Rabs_right; [done| apply Rle_ge; apply pos_norm].
have Hpser:=  GP_infinite (mynorm X) H4.
red in Hpser; red in Hpser.

move=> eps He; case (Hpser eps He)=> N Hn0; exists N => n Hn.
have H5:= Hn0 n Hn.
rewrite /R_dist in H5 |-*.
apply Rle_lt_trans with 
 (Rabs (sum_f_R0 (fun n : nat => 1 * mynorm X ^ n) n - / (1 - mynorm X)) ).
right.
congr (Rabs _).
congr ( _ - _).
rewrite -echiv_sums //=.
elim n => //=.
by simp_R.
move=> n0 Ihn. rewrite Ihn; simp_R.
congr ( _ + _ ); congr ( _ * _).
elim n0 => //=; move=> n1 Ihn1; simp_R; by congr ( _ * _ ).
by rewrite /Rdiv Rmult_1_l.
done.
Qed.



End mat0.


Implicit Arguments dif_m [p].







 