Require Import ssreflect ssrbool funs eqtype ssrnat seq fintype.
Require Import tuple paths connect div .
Require Import groups group_perm zp ssralg ssrbig matrix.
Require Import structRp min_max.
Require Import Reals.
Require Import Fourier.
Require Import Ranalysis.
Require Import Classical.
Require Import ClassicalChoice.
Require Import vectors seqRp matrixNorm.
Require Import Max.

(*
Ltac simp_R:=
match goal with 
| |- (_ = _ )=> rewrite /(@Ring.opp ring_add_grp ) /=
                        /(@Ring.add ring_add_grp ) /= 
                        /(@Ring.mul ring_base ) /=
                       /0 /= /1 /=; ring_simplify
| _ =>  rewrite /(@Ring.opp ring_add_grp ) /=
                        /(@Ring.add ring_add_grp ) /= 
                        /(@Ring.mul ring_base ) /=
                       /0 /= /1 /=
end.
*)

Section mat.
Variable p: pos_nat.
Lemma Hp: (O<p)%nat.
Proof.
have Hp1:= pos_natP p.
move/leP:Hp1=>Hp1.
omega.
Qed.


Notation "'MR_' p" := (matrix real_field p p)(at level 10).

Definition sum_abs (A: MR_(p)) i:R := \sum_(j) (Rabs (A i j)).
Definition sum_abs2 (v: (Rvec p)) :R := \sum_(j) (Rabs (v j)).

Definition norm_m (A: MR_(p)):R:= \big[Rmax/R0]_(i) (sum_abs A i).

(*Definition norm_m2 (A: MR_(p)):R:=\big[Rmax/R0]_(i) (sum_abs2 (\vec_(j)  (A i j))).
*)
(* interpret vectors as column matrices
Definition matr_of_vec (v:(Rvec p)): matrix real_ring p 1 :=
\matrix_(i,j) (v i).

Definition vec_of_matr (m: matrix real_ring p 1) : (Rvec p):=
\vec_(i) (m i ord0).

Coercion matr_of_vec: (Rvec p) >-> matrix .
*)

(*define separatly the multiplication of matrices and vectors*)
Notation "\vec_ ( i ) E" :=
  (@vec_of_fun p  (fun i  => E)) (at level 25).

Definition mult_mv (m:MR_(p)) (v:(Rvec p)) :=
\vec_(i) (\sum_(j) m i j* v j). 

Lemma sum_abs_pos: forall a i, R0 <= sum_abs a i.
Proof.
move =>a i.
rewrite /sum_abs .
unlock reducebig.
move : (I_not_void p Hp).
elim: (index_enum I_(p)) =>//=.
move => x s IH1 _.
apply Rplus_le_le_0_compat.
apply Rabs_pos.
case A: (seq0 != s).
exact: IH1 A.
move/eqP: A =>A
; rewrite -A //=.
rewrite /0 //=; fourier. (*even the 0 (zero) is hidden..*)
Qed.

Lemma eq_sum_abs:
forall a i, sum_abs a i = Rabs (sum_abs a i).
Proof.
move => *; rewrite Rabs_right; [done| ].
apply Rle_ge; apply sum_abs_pos.
Qed.

Lemma abs_sum_le:
forall (v:(Rvec p)), Rabs (\sum_(i) (v i)) <= \sum_(i) (Rabs (v i)). 
Proof.
move => v.
unlock reducebig.
move : (I_not_void p Hp); rewrite /index_enum.
elim : (enum I_(p)) (uniq_enum I_(p)) => //=.
move => x s IH IHun _ .
apply Rle_trans with (Rabs (v x) + Rabs (foldr (fun (i : ordinal p) (x0 : R) => v i + x0) 0 s) ).
apply Rabs_triang.
apply Rplus_le_compat_l.
move/andP : IHun => [In Iu].
case C:(seq0!=s).
apply (IH Iu C).
move/eqP: C=>C.
rewrite -C //= Rabs_R0; apply Rle_refl.
Qed.

Lemma abs_sum_le_f:
forall (f:I_(p) ->R) , Rabs (\sum_(i) (f i)) <= \sum_(i) (Rabs (f i)). 
Proof.
move => v.
unlock reducebig.
move : (I_not_void p Hp );rewrite /index_enum.
elim : (enum I_(p)) (uniq_enum I_(p)) => //=.
move => x s IH IHun _ .
apply Rle_trans with (Rabs (v x) + Rabs (foldr (fun (i : ordinal p) (x0 : R) => v i + x0) 0 s) ).
apply Rabs_triang.
apply Rplus_le_compat_l.
move/andP : IHun => [In Iu].
case C:(seq0!=s).
apply (IH Iu C).
move/eqP: C=>C.
rewrite -C //= Rabs_R0; apply Rle_refl.
Qed.

Lemma pres_ineq1:
forall (f r:R)( v:(Rvec p)),(forall j,  Rabs (v j)<= r )-> ( 0 <= f  )->
\sum_(i) (Rabs (v i)* f )<= \sum_(i<p)( r*f ).
Proof.
move =>f v r Hv Hf.
unlock reducebig.
rewrite /index_enum.
elim : (enum I_(p))  => //=; first by apply Rle_refl.
move => x s IH .
apply Rplus_le_compat; last by done.
apply Rmult_le_compat_r; [by apply Hf|by apply Hv].
Qed.

Lemma pres_ineq:
forall (f:I_(p)->R)( v:(Rvec p)) r,(forall j,  Rabs (v j)<= r )-> (forall x, 0 <= f x )->
\sum_(i) (Rabs (v i)* f i)<= \sum_(i)( r*f i).
Proof.
move =>f v r Hv Hf.
unlock reducebig.
rewrite /index_enum.
elim : (enum I_(p))  => //=; first by apply Rle_refl.
move => x s IH .
apply Rplus_le_compat; last by done.
apply Rmult_le_compat_r; [by apply Hf|by apply Hv].
Qed.

(*
Lemma eq_def_norm:
forall a, norm_m a = norm_m2 a.
Proof.
move =>a.
rewrite /norm_m /norm_m2 /sum_abs /sum_abs2.
apply eq_bigr.
rewrite /dfequal => x _.
apply eq_bigr.
rewrite /dfequal => y _.
by rewrite vxK.
Qed.
*)

Lemma desc_norm_m:forall a r, norm_m a = Rmax (sum_abs a r)
    (\big[Rmax/R0]_(k<p|k!=r) (sum_abs a k)).
Proof.
move=>a r; rewrite /norm_m.
have H:= sum_abs_pos a.
by apply sxbigD1 with (P:=(fun _  :I_(p)=> true)) (j:=r) (F:= (fun t => sum_abs a t)).
Qed.

Lemma normM_ex:
forall m, exists k, norm_m m = sum_abs m k.
Proof.
move => m.
rewrite /norm_m.
apply (max_exists p Hp).
apply sum_abs_pos.
Qed.

Lemma normM_pos:
forall m, 0 <= norm_m m.
Proof.
move => m.
elim (normM_ex m) => k Hrel.
rewrite Hrel.
apply sum_abs_pos.
Qed.




Lemma normM_max:
forall m i, sum_abs m i <= norm_m m. 
Proof.
move => m i.
rewrite (desc_norm_m m i); apply RmaxLess1.
Qed.

(*use the operations on matrices from matrix.v;
  ? define the multiplication between a matrix and a vector
or
  ? define a coercion between matrix (p, 1) and vectors

it is not automatic that a row or a column of a matrix is a vector....
*)






Lemma ineq_norm_mv:
forall m v,
norm (mult_mv m v) <= norm_m m * norm v. 
Proof.

move => m v.
elim (norm_rew Hp (mult_mv m v)) => k1 Heq1.
rewrite Heq1.
rewrite /mult_mv vxK.

apply Rle_trans with (\sum_(j) (Rabs (m k1 j * v j))).
apply abs_sum_le_f.

have Heq_abs: (fun j =>Rabs (m k1 j * v j))  =1 (fun j => Rabs (v j)* Rabs (m k1 j)). 
rewrite /eqfun => i;rewrite Rmult_comm; apply Rabs_mult.
have Heq_prf: \sum_(j) Rabs (m k1 j * v j)  = \sum_(j) (Rabs (v j)*Rabs (m k1 j)).
apply eq_bigr.
by rewrite /dfequal => i _; rewrite  //=  Rabs_mult.
rewrite Heq_prf.
elim (norm_rew Hp v) => k2 Heq2.
elim (normM_ex m) => k3 Heq3.
rewrite Heq2 Heq3 Rmult_comm.
apply Rle_trans with (\sum_(j) (Rabs (v k2)*Rabs (m k1 j))).
apply pres_ineq. 
rewrite -Heq2.
apply norm_max.
move => x; apply Rabs_pos.
rewrite /sum_abs -big_distrr /=.
apply Rmult_le_compat_l .
apply Rabs_pos.

have H7: \sum_(i < p) Rabs (m k1 i)=(sum_abs m k1).
rewrite /sum_abs //=.
have H8: \sum_(i < p) Rabs (m k3 i)=(sum_abs m k3).
rewrite /sum_abs //=.
rewrite H7 H8 -Heq3.
apply normM_max.
Qed.

Lemma ineq_norm_m:
forall a b,
norm_m (a *m b) <= norm_m a*norm_m b.
Proof.
move =>a b.
rewrite {1}/norm_m /mulmx /sum_abs.
have H1: \big[Rmax/0]_(i) (\sum_(j) Rabs
                             ((\matrix_(i0, k) (\sum_(j0) a i0 j0 * b j0 k))
                                i j)) = 
\big[Rmax/0]_(i) (\sum_(j) Rabs (\sum_(s) a i s*b s j)).
apply eq_bigr => i _.
apply eq_bigr => j _.
by rewrite mxK.
rewrite H1.
apply Rle_trans with (\big[Rmax/0]_(i) (\sum_(j) (\sum_(s) (Rabs (a i s * b s j))))).
apply (Rmax_le_f_le ) => i.
apply (sum_le_f_le ) => j.
apply abs_sum_le_f.
have H2: \big[Rmax/0]_(i) (\sum_(j) \sum_(s) Rabs (a i s * b s j))  = 
               \big[Rmax/0]_(i) (\sum_(j) \sum_(s) Rabs (a i s )*Rabs( b s j)) .
apply eq_bigr =>i _; apply eq_bigr => j _;apply eq_bigr => s _; apply Rabs_mult.
rewrite H2.
have H3: \big[Rmax/0]_(i) (\sum_(j) \sum_(s) Rabs (a i s) * Rabs (b s j)) =
\big[Rmax/0]_(i) (\sum_(s) Rabs (a i s)*\sum_(j)  Rabs (b s j)) .
have H4: \big[Rmax/0]_(i) (\sum_(j) \sum_(s) Rabs (a i s) * Rabs (b s j)) =
\big[Rmax/0]_(i) (\sum_(s) \sum_(j) Rabs (a i s) * Rabs (b s j))  .
apply eq_bigr => i _.
apply: exchange_big.
rewrite H4.
apply eq_bigr => i _; apply eq_bigr => s _.
rewrite -big_distrr //=.
rewrite H3.
apply Rle_trans with (\big[Rmax/0]_(i) (\sum_(s) Rabs (a i s) *norm_m b)).
apply (Rmax_le_f_le ) => i.
apply (sum_le_f_le ) => s.
apply Rmult_le_compat_l; first by apply Rabs_pos.
fold (sum_abs b s).
 apply normM_max.
have H5: \big[Rmax/0]_(i) (\sum_(s) Rabs (a i s) * norm_m b)  = 
(\big[Rmax/0]_(i) (\sum_(s) Rabs (a i s) ))* norm_m b.
rewrite Rmult_comm.
have H6 : \big[Rmax/0]_(i) (\sum_(s) Rabs (a i s) * norm_m b) =
\big[Rmax/0]_(i) (\sum_(s) norm_m b *Rabs (a i s) ) .
apply eq_bigr => i _;
apply eq_bigr => s _; by rewrite Rmult_comm.
rewrite H6.
rewrite distr_mult_max.
apply eq_bigr => i _.
by rewrite -big_distrr.
apply normM_pos.
rewrite H5.
have H7: \big[Rmax/0]_(i) (\sum_(s) Rabs (a i s)) = norm_m a.
by rewrite /norm_m /sum_abs.
rewrite H7.
apply Rle_refl.
Qed.


Lemma normM_plus_le:
forall a b, norm_m (a +m b) <= norm_m a+ norm_m b.
Proof.
move => a b.
rewrite {1}/norm_m /addmx /sum_abs.
have H1:
\big[Rmax/0]_(i) (\sum_(j) Rabs ((\matrix_(i0, j0) (a i0 j0 + b i0 j0)) i j))
= \big[Rmax/0]_(i) (\sum_(j) Rabs (a i j + b i j)).
apply eq_bigr => i _.
apply eq_bigr => j _.
by rewrite mxK.
rewrite H1.
apply Rle_trans with (\big[Rmax/0]_(i) (\sum_(j) (Rabs (a i j) +Rabs( b i j)))).
apply (Rmax_le_f_le )=> i.
apply (sum_le_f_le ) => s.
apply Rabs_triang.
have H2: \big[Rmax/0]_(i) (\sum_(j) (Rabs (a i j) + Rabs (b i j))) =
\big[Rmax/0]_(i) (\sum_(j) Rabs (a i j) +\sum_(j) Rabs (b i j)) .
apply eq_bigr => i _.
rewrite big_split //=.
rewrite H2.
apply Rle_trans with (\big[Rmax/0]_(i) (\sum_(j) Rabs (a i j)) +
\big[Rmax/0]_(i) ( \sum_(j) Rabs (b i j)) ).
apply (distr_plus_max Hp).
move => i; fold (sum_abs b i); apply sum_abs_pos.
move => i;  fold (sum_abs a i); apply sum_abs_pos.
rewrite /norm_m /sum_abs.
apply Rle_refl.
Qed.

Lemma hom_pos_normM:
forall m a, norm_m (scalemx a m) = Rabs a * norm_m m.
Proof.
move => v a.
have Hnec: norm_m (scalemx a v) <= Rabs a*norm_m v.
elim (normM_ex (scalemx a v)); intros k Heq.
rewrite Heq /scalemx /sum_abs.
have H: \sum_(j) Rabs ((\matrix_(i, j0) (a * v i j0)) k j) = 
\sum_(j) Rabs a*Rabs ((\matrix_(i, j0) ( v i j0)) k j).
by apply: eq_bigr => j _; rewrite !mxK Rabs_mult.
rewrite H -big_distrr //=.
apply Rmult_le_compat_l; [apply Rabs_pos| ].
have H1: \sum_(i < p) Rabs ((\matrix_(i0, j0) v i0 j0) k i)= (sum_abs  v k).
by rewrite /sum_abs //=; apply: eq_bigr => i _; rewrite mxK.
rewrite H1.
apply: normM_max.
have H2: Rabs a * norm_m v <= norm_m (a *s v).
elim (normM_ex v) => k Heq.
rewrite Heq eq_sum_abs -Rabs_mult /sum_abs.
have H: Rabs (a * (\sum_(j) Rabs (v k j))) = Rabs ((\sum_(j) a*Rabs (v k j))).
congr (Rabs _).
rewrite -big_distrr //=.
rewrite H.
apply Rle_trans with (\sum_(j) Rabs (a * Rabs (v k j))).
apply abs_sum_le_f.
have H3: \sum_(j) Rabs (a * Rabs (v k j)) = sum_abs (a *s v) k.
rewrite /sum_abs /scalemx.
by apply eq_bigr => i _; rewrite mxK Rabs_mult Rabs_Rabsolu Rabs_mult.
rewrite H3.
apply normM_max.
by apply Rle_antisym.
Qed.




Lemma zeroM_0:
norm_m (\0_(p))= 0.
Proof.
elim (normM_ex \0_(p)) => k Hk.
rewrite Hk.
rewrite /sum_abs .
unlock reducebig.
elim (index_enum I_(p)) => //=.
move => x s IH.
 rewrite {1}/null_mx  mxK //= Rabs_R0 IH   /0 //=.
by simp_R.
Qed.

Lemma norm_unitM:
norm_m \1_(p) = 1.
Proof.
elim (normM_ex \1_(p)) => k Hk.
rewrite Hk /sum_abs /unit_mx /scalar_mx (bigD1 k) //= !mxK eq_refl Rabs_R1.
have H: \sum_(i < p | i != k) Rabs ((\matrix_(i0, j) (if i0 == j then 1 else 0)) k i) 
=0.
unlock reducebig.
elim (index_enum I_(p)) => //=.
move => x s IH.
case X:(x!=k).
rewrite IH mxK .
have H: (k==x)=false.
apply/eqP.
move/eqP:X=>X.
rewrite /not in X |-*=> H.
have H1: x = k by apply sym_eq.
by apply X.
 rewrite H Rabs_R0.
 rewrite (@Ring.add0r ring_add_grp) //= .
apply IH.
 rewrite H. 
by simp_R.
Qed.

Delimit Scope ring_scope with Ri.

Lemma mult_mv_assoc:
forall a b v, mult_mv a (mult_mv b v) = mult_mv ( a *m b) v.
Proof.
move => a b v.
rewrite /mult_mv /mulmx .
rewrite -ext_eq => i.
rewrite !vxK.
transitivity (\sum_(j)  (\vec_ (i0)(\sum_(j0) a i j * b i0 j0 * v j0)) j ).
apply eq_bigr => j _.
rewrite !vxK.
have H: (a i j * (\sum_(j0) b j j0 * v j0))%Ri = a i j * (\sum_(j0) b j j0 * v j0).
by simp_R.
rewrite -H.
rewrite big_distrr //=; simp_R.
by apply: eq_bigr => k _; rewrite Rmult_assoc.
transitivity (\sum_(j) (\sum_(j0) (\vec_ (i0) (a i j * b i0 j0 * v j0)) j)) .
apply: eq_bigr => j _.
rewrite vxK.
apply: eq_bigr => j0 _.
by rewrite vxK.
transitivity (\sum_(j)  (\sum_(j0) (\matrix_(i0, k) (a i0 j0 * b j0 k)) i j) * v j).
rewrite exchange_big //=.
apply: eq_bigr => j _.
rewrite Rmult_comm.
have H : v j * (\sum_(j0 < p) (\matrix_(i0, k) (a i0 j0 * b j0 k)) i j)=
(v j * (\sum_(j0 < p) (\matrix_(i0, k) (a i0 j0 * b j0 k)) i j))%Ri.
 by simp_R.
rewrite H.
rewrite  big_distrr //=; simp_R.
apply : eq_bigr => j0 _.
by rewrite vxK mxK  //=; ring_simplify.
apply : eq_bigr => j _.
rewrite !mxK.
have H : (\sum_(j0) (a i j0 * b j0 j)%Ri) * v j = ((\sum_(j0) (a i j0 * b j0 j)%Ri) * v j)%Ri.
by simp_R.
rewrite H.
rewrite big_distrl //=.
have H1: (\sum_(j0 < p) (\matrix_(i0, k) (a i0 j0 * b j0 k)) i j) * v j = 
((\sum_(j0 < p) (\matrix_(i0, k) (a i0 j0 * b j0 k)) i j) * v j)%Ri by simp_R.
rewrite H1.
rewrite big_distrl //=; simp_R.
apply: eq_bigr => k _.
by rewrite mxK.
Qed.


 Lemma opp_distr:
forall m v,
mult_mv m (opp_v v) = opp_v (mult_mv m v).
Proof.
move =>m v.
rewrite /mult_mv /opp_v.
rewrite -ext_eq =>  i; rewrite //= !vxK  .
have H:  - (\sum_(j < p) m i j * v j) = (- (\sum_(j < p) m i j * v j))%Ri by simp_R.
rewrite H -sum_opp ; simp_R.
by apply: eq_bigr => j _; rewrite !vxK  //=; ring_simplify.
Qed.

(* unit matrix is a neutral element for matrix vector multiplication*)
Lemma unitM_el_neu:
forall v,
mult_mv \1 v = v.
Proof.
move=> v.
rewrite /mult_mv.
rewrite -ext_eq => i; rewrite !vxK (bigD1 i) //= mxK .
rewrite big1 ?simp //=.
rewrite set11; by simp_R.
move => j; rewrite //= ; move=> H.
by rewrite mxK eq_sym (negbET H); simp_R.
Qed.

(*
Definition dif_m  (A B:MR_(p)) :=
\matrix_(i, j) (A i j - B i j:structRp.real_field).
*)
Notation "A -m B" := (dif_m A B) (at level 50, format "A  -m  B").

Lemma distr_m_dif:
forall (a b d: MR_(p)),  (a *m b) -m (a *m d) =  a *m (b -m d).
Proof.
move=>  A1 A2 B; apply/matrixP=> i k; rewrite !mxK.
replace (\sum_(j) (A1 i j * A2 j k)%Ri - \sum_(j) (A1 i j * B j k)%Ri ) with 
((\sum_(j) (A1 i j * A2 j k) - \sum_(j) (A1 i j * B j k))%Ri) ;[| by simp_R ].
rewrite -sum_opp -big_split /=.
by apply: eq_bigr => j _; rewrite mxK; simp_R.
Qed.

Lemma mult_v0:
forall m, mult_mv m (vect0 p) = vect0 p.
Proof.
move => m.
rewrite /mult_mv /vect0 -ext_eq  =>i;
rewrite !vxK.
have H2: \sum_(j) m i j * vec_of_fun (fun _ : I_(p) => 0) j =
\sum_(j)  m i j * 0 .
by apply eq_bigr => j _; rewrite !vxK; simp_R.
by rewrite H2 -big_distrl; simp_R.
Qed.

Lemma distr_opp_mv:
forall m1 m2 v,
dif_v (mult_mv m1 v) (mult_mv m2 v) = mult_mv (m1 -m m2) v.
Proof.
move=> m1 m2 v.
rewrite -ext_eq /dif_v /dif_m /mult_mv  => i; rewrite !vxK .
rewrite /Rminus.
replace (\sum_(j) m1 i j * v j + - (\sum_(j) m2 i j * v j)) with
((\sum_(j) m1 i j * v j + - (\sum_(j) m2 i j * v j))%Ri);[| by simp_R].
rewrite -sum_split_sub.
by apply eq_bigr => j _; rewrite !mxK; simp_R.
Qed.

Lemma distr_opp_vm:
forall m v1 v2,
dif_v (mult_mv m v1) (mult_mv m v2) = mult_mv m (dif_v v1 v2).
Proof.
move => m u v.
rewrite -ext_eq /dif_v  /mult_mv  => i; rewrite !vxK.
rewrite /Rminus.
replace (\sum_(j) m i j * u j + - (\sum_(j) m i j * v j)) with
((\sum_(j) m i j * u j + - (\sum_(j) m i j * v j))%Ri); [|by simp_R].
rewrite -sum_split_sub.
by apply eq_bigr => j _; rewrite !vxK; simp_R.
Qed.

Lemma canonical1:
 forall (A:MR_(p)) i j, Rabs (A i j) <= norm_m A.
Proof.
move => a i j.
rewrite /norm_m.
apply Rle_trans with (sum_abs a i);[|apply max_max; apply sum_abs_pos].
rewrite /sum_abs.
rewrite (bigD1 j ) //=.
rewrite -{1}(Rplus_0_r (Rabs (a i j))).
apply Rplus_le_compat_l.
unlock reducebig; elim (index_enum I_(p)) => //=;[by right|].
move=> x s IH.
case (x!=j);[|done].
rewrite -{1}(Rplus_0_r R0).
apply Rplus_le_compat;[apply Rabs_pos|done].
Qed.


Lemma canonical2:
forall (A B:MR_(p)), (forall i j, Rabs (A i j) <= Rabs (B i j)) 
                                         ->norm_m A <= norm_m B.
Proof.
move=> a b H.
rewrite /norm_m; apply (Rmax_le_f_le ) => i.
by rewrite /sum_abs; apply (sum_le_f_le ) => j.
Qed.


Canonical Structure the_norm :=
MNorm p norm_m normM_pos hom_pos_normM normM_plus_le ineq_norm_m
canonical1 canonical2.

Lemma  matr_inv_norm:
forall a: MR_(p), norm_m  (\1 -m a) < 1 ->  \det a <>0.
Proof.
move=> a Ha.
apply (cor1_th5 p _ norm_unitM).
by rewrite //=.
Qed.

Notation "A ^-1m" := (invmx  A) (at level 10, format "A ^-1m") .

Lemma norm_inv_ser:
forall a: MR_(p),
norm_m a <1->
norm_m (( \1 -m a)^-1m) <= 1/(1 - norm_m a).
Proof.
move=> a H.
apply (norm_inv_ser0 p _  norm_unitM _ H ).
exact norm_unitM.
Qed.


End mat.

Implicit Arguments mult_mv [p].
Implicit Arguments norm_m [p].
Implicit Arguments abs_sum_le_f [p].
