(*Require Import ssreflect ssrbool funs eqtype ssrnat seq fintype.
Require Import tuple paths connect div .
Require Import groups group_perm zp ssralg ssrbig matrix.
*)
Require Import ssreflect eqtype fintype ssrbool ssralg.
Require Import Reals.
Require Import Fourier.
Require Import Ranalysis.
Require Import Classical.


Open Scope R_scope.



Section real_eqtype.

Variable p: nat.
Hypothesis Hp: (O<p)%nat.

Definition eqr (r1 r2 : R) : bool := 
let H:=(total_order_T r1 r2) in
match  H with
| inleft (left _) => false
| inleft (right _) => true
| inright _ => false
end.

Lemma eqrP : reflect_eq eqr.
Proof.
move=>r1 r2; apply: (iffP idP) => [|<-].
rewrite /eqr.
case (total_order_T r1 r2) =>//=.
case=>//=.
rewrite /eqr.
case (total_order_T r1 r1) =>//=; try tauto.
case=>//=; auto.
have H:= Rlt_irrefl r1.
done.
rewrite /Rgt.
have H:= Rlt_irrefl r1.
done.
Qed.

End real_eqtype.

Canonical Structure eR := EqType (@eqrP).
(*we defined an eq_type for the real numbers*)

Lemma R_irrelevance : forall (x y : R) (E E' : x = y), E = E'.
Proof. exact: eq_irrelevance. Qed.

Import Ring.

(*we define a commutative ring structure for the real numbers, 
so that we may use the matrices in ssreflect*)

Lemma eRplus_assoc : associative (Rplus).
by move=> *; ring_simplify.
Qed.
Lemma eRplus_com : commutative (Rplus).
by move=> *; ring_simplify.
Qed.
Lemma eR0plus : left_unit R0 Rplus.
by move=> *; ring_simplify.
Qed.
Lemma eRNplus : left_inverse R0 Ropp Rplus.
by move=> *; ring_simplify.
Qed.

Canonical Structure ring_add_grp :=
 AdditiveGroup eRplus_assoc eRplus_com eR0plus eRNplus.


Lemma eRmult_assoc : associative (Rmult).
by move=> *; ring_simplify.
Qed.
Lemma left_unitRmult: left_unit R1 Rmult.
by move=> *; ring_simplify.
Qed.
Lemma right_unitRmult : right_unit R1 Rmult.
by move=> *; ring_simplify.
Qed.
Lemma LDRAM : left_distributive Rmult Rplus.
by move=> *; ring_simplify.
Qed.
Lemma RDRAM : right_distributive Rmult Rplus.
by move=> *; ring_simplify.
Qed.
Lemma R1_neq_0 : R1 <> R0.
by exact: R1_neq_R0.
Qed.

Canonical Structure ring_base :=
 Basic eRmult_assoc left_unitRmult right_unitRmult LDRAM RDRAM
    R1_neq_0.

Lemma Rmult_com : commutative Rmult.
by move=> *; ring_simplify.
Qed.

Canonical Structure real_ring := Commutative Rmult_com.

Import Field.

Lemma Rinv_inv: forall x:R, x <> 0 -> (Rinv x) * x = 1.
Proof.
move=> x Hx; rewrite /mul //= /1 //=; by field.
Qed.


Canonical Structure real_field :=
Field.Field Rinv_inv.
