(*
Require Import ssreflect ssrbool funs eqtype ssrnat seq fintype.
Require Import tuple paths connect div .
Require Import groups group_perm zp ssralg ssrbig matrix.
*)
Require Import ssreflect eqtype fintype funs ssrbig.
Require Import structRp min_max.
Require Import Reals.
Require Import Fourier.
Require Import Ranalysis.
Require Import Classical.


Section vectors.

(*definition and properties for vectors of length p*)
Variable p: nat.
Hypothesis Hp: (O<p)%nat.

Print Canonical Projections.

CoInductive Rvec : Type := 
 Vect of fgraphType I_(p) eR.
  
Definition vval v := match v with Vect g => g end.

Lemma can_vval : cancel vval Vect.
Proof. by rewrite /cancel; case => /=. Qed.

Lemma vval_inj : injective vval. 
Proof. exact: can_inj can_vval. Qed.

Canonical Structure vec_eqType := EqType (can_eq can_vval).

Definition vec_of_fun F :=
  locked Vect (fgraph_of_fun (fun i => F i)).

Definition fun_of_vec := locked (fun A i  => vval A i).

Coercion fun_of_vec : Rvec >-> Funclass.

Lemma vxK : forall F, vec_of_fun F =1 F.
Proof. by unlock vec_of_fun fun_of_vec => F i ; rewrite /= g2f. Qed.

Lemma ext_eq : forall A B : Rvec, A =1 B <-> A = B.
Proof.
unlock fun_of_vec => [] [A] [B]; split=> [/= eqAB | -> //].
congr Vect; apply/fgraphP=> [] [i j]; exact: eqAB.
Qed.
(*extensional equality on vectors is a lemma and not an axiom*)

Lemma ext_eq1: forall u v : Rvec, u =1 v -> u = v.
Proof.
by move => u v Huv; rewrite -ext_eq.
Qed.

Reserved Notation "\vec_ ( i ) E"  (at level 25).

Notation "\vec_ ( i ) E" :=
  (@vec_of_fun  (fun i  => E)) (at level 25).


(*operations on vectors*)

Definition add_v (v1 v2:Rvec):= \vec_(i) (v1 i + v2 i).
Definition mult_sv (a:R) (v:Rvec):= \vec_(i) (a* v i).
Definition dif_v (v1 v2:Rvec):= \vec_(i) (v1 i- v2 i).
Definition opp_v (v:Rvec):=\vec_(i) ( -v i).

(*canonical base*)
Definition vect0:= \vec_(i) R0.
Definition base_v k:=\vec_(i)(if k==i then R1 else R0).


Notation "v1 +^ v2" := (add_v v1 v2) (at level 50).
Notation "a *^ v" := (mult_sv a v) (at level 40).
Notation "v1 -^ v2" := (dif_v v1 v2) (at level 50).
Notation "-^ v" := (opp_v v) (at level 35).

Lemma mult_add: 
forall a v1 v2 ,
a *^ (v1 +^ v2) = a*^ v1 +^ a*^ v2.
Proof.
move=> a v1 v2.
rewrite /add_v /mult_sv .
apply/ext_eq => i.
rewrite !vxK.
by ring_simplify.
Qed.

(*definition and properties for a norm*)
Definition norm (v:Rvec):= \big[Rmax/R0]_(i<p) (Rabs (v i)).

Lemma norm_elem_eq:
forall (u v:Rvec),
(forall i, Rabs (u i)=Rabs (v i))-> norm u = norm v.
Proof.
move => u v Huvi;
by apply: eq_bigr=> i _; rewrite Huvi.
Qed.


Lemma elem_le_norm_le:
forall (v1 v2:Rvec),
(forall i, Rabs (v1 i) <= Rabs (v2 i))-> norm v1 <= norm v2.
Proof.
move=> v1 v2 Hrabs.
apply: big_rel; first by right.
move=> a x b y Hax Hby.
by rewrite /Rmax; case (Rle_dec a b) => Hab; case (Rle_dec x y) => Hxy; 
try fourier; apply Rle_trans with y; try fourier; left; apply Rnot_le_lt .
move=> i _; apply Hrabs.
Qed.


Lemma norm_pos: forall v,R0<=norm v.
Proof.
move => v.
apply: big_prop; [by right|
move=> x y Hx Hy; apply/Rmax_Rle; by left|
move=> i _; apply Rabs_pos].
Qed.


Lemma desc_norm:forall v j, norm v = Rmax (Rabs (v j))
    (\big[Rmax/R0]_(i<p|i!=j) (Rabs (v i))).
Proof.
move=>v j; rewrite /norm.
apply sxbigD1 with (P:=(fun _  :I_(p)=> true)) (j:=j) (F:= (fun t => Rabs (v t))).
move => i; apply : Rabs_pos.
done.
Qed.


Lemma norm_max: forall (v:Rvec) i, Rabs (v i) <= norm v.
Proof.
move=> v i; rewrite (desc_norm v i); apply RmaxLess1.
Qed.

Lemma norm_rew:
forall v,exists k, norm v = Rabs (v k). 
Proof.
move => v.
rewrite /norm.
apply (max_exists p Hp).
move => i; apply Rabs_pos.
Qed.

Lemma norm_tri: forall (v1 v2:Rvec), norm (add_v v1 v2)<=norm v1 + norm v2.
Proof.
move=> v1 v2.
elim (norm_rew (add_v v1 v2))=> k Heq.
rewrite Heq /add_v.
apply Rle_trans with (Rabs (v1 k)+Rabs (v2 k)).
rewrite vxK.
apply Rabs_triang.
apply Rplus_le_compat ; apply norm_max.
Qed.


Lemma  norm_hom_pos:
forall v (a:R),
norm (mult_sv a v) = Rabs a*norm v.
Proof.
move => v a.
have Hnec: norm (mult_sv a v) <= Rabs a*norm v.
elim (norm_rew (mult_sv a v)); intros k Heq.
rewrite Heq /mult_sv vxK Rabs_mult.
apply Rmult_le_compat_l; [apply Rabs_pos| apply norm_max].
have H2:Rabs a*norm v<= norm (mult_sv a v).
elim (norm_rew v); intros k  Heq.
rewrite Heq  -Rabs_mult.
have J: (a* v k)%R = (a*^ v) k.
by rewrite /mult_sv vxK.
rewrite J.
apply (norm_max (mult_sv a v) k ).
by apply Rle_antisym.
Qed.

Lemma Rabs0_R0:
forall r,
Rabs r = R0 <-> r = R0.
Proof.
move => r. 
split =>H.
split_Rabs; apply Rle_antisym; fourier.
rewrite H;apply Rabs_R0.
Qed.


Lemma norm0:
forall v, norm v = R0 <-> v = vect0.
Proof.
move => v.
split; move => H.
rewrite -ext_eq=> i; rewrite /vect0 vxK.
have Hineq: Rabs(v i) <= R0.
apply Rle_trans with (norm v).
apply norm_max.
unfold Rle; right; assumption.
have Heq: Rabs (v i)=R0 .
 apply Rle_antisym; [assumption|apply Rabs_pos].
by rewrite -Rabs0_R0.
elim (norm_rew v)=> k Heq.
rewrite Heq.
rewrite H /vect0 vxK; apply Rabs_R0.
Qed.

Lemma dif_eq_vect0:
forall (u v:Rvec),
dif_v u v = vect0 <-> u = v.
Proof.
move => u v.
split => H.
rewrite /vect0  in H.
rewrite -ext_eq => i.
have Hdif: dif_v u v i = R0 by rewrite (H ) vxK.
rewrite /dif_v vxK in Hdif.
by apply Rminus_diag_uniq.
rewrite H.
rewrite - ext_eq => i.
rewrite /vect0  /dif_v !vxK.
by ring_simplify.
Qed.



Lemma norm0_eq:
forall u v, norm (dif_v u v) = R0 <-> u = v.
Proof.
move => u v.
split=> H.
by rewrite (norm0 (dif_v u v)) dif_eq_vect0 in H.
by rewrite (norm0 (dif_v u v)) dif_eq_vect0.
Qed.

Lemma Rabs_base:
forall k, Rabs (base_v k k) = R1 .
Proof.
move =>k.
rewrite /base_v vxK.
rewrite eq_refl.
apply Rabs_R1.
Qed.

Lemma norm_base_v:
forall k, norm (base_v k) = R1.
Proof.
move => k.
rewrite -(Rabs_base k).
elim (norm_rew (base_v k)) => k0 Heq.
case C: (k==k0).
by rewrite -(eqP C) in Heq.
rewrite {2}/base_v vxK in Heq.
rewrite C Rabs_R0 in Heq.
have Hco:R1<=R0 by rewrite -Heq -(Rabs_base k); by apply norm_max.
have Hco2: ~(R1<=R0) by fourier.
contradiction.
Qed.


Lemma ex_v_norm:
forall r, (R0<r)%R -> exists x, norm x = r.
Proof.
move => r Hpos.
exists (mult_sv r (base_v (I0 p Hp)  )).
rewrite norm_hom_pos norm_base_v Rabs_right;
fourier.
Qed.

Lemma el_div_norm_leq:
forall (v:Rvec) x, 0<norm v ->Rabs (v x) /norm v <= 1.
Proof.
move => v x Hv.
rewrite /Rdiv.
have H:= norm_max v x.
apply Rmult_le_reg_l with (norm v); first by done.
rewrite Rmult_comm Rmult_assoc Rinv_l; last by apply Rgt_not_eq; apply Hv.
ring_simplify; apply: norm_max.
Qed.



(*definition and properties for distance*)

Definition dist_Rp (x y :Rvec) :R := norm (dif_v x y).

Lemma dist_pos:
forall v u, R0<= dist_Rp v u.
Proof.
move => v u.
rewrite /dist_Rp.
apply norm_pos.
Qed.

Lemma dist_sym:
forall u v, dist_Rp u v = dist_Rp v u.
Proof.
move=> u v.
rewrite /dist_Rp.
have Heqdif:forall i, Rabs ((dif_v u v) i) = Rabs ((dif_v v u) i).
move => i.
by rewrite /dif_v !vxK Rabs_minus_sym.
by apply norm_elem_eq.
Qed.

Lemma dist_tri:
forall u v t,
dist_Rp u v <= dist_Rp u t + dist_Rp t v.
Proof.
move => u v t.
rewrite /dist_Rp.
have Heq: forall i, (dif_v u v) i = (add_v (dif_v u t) (dif_v t v) )i.
move=>i; rewrite /dif_v /add_v !vxK.
by ring_simplify.
have Hn: norm (dif_v u v) = norm (add_v (dif_v u t) (dif_v t v)).
 apply  norm_elem_eq.
move => i;by  rewrite (Heq i) .
rewrite Hn.
apply norm_tri.
Qed.

Lemma dist_refl:
forall u v,
dist_Rp u v = R0 <-> u = v.
Proof.
move => u v.
 rewrite /dist_Rp.
split => H.
by  rewrite -norm0_eq.
by rewrite norm0_eq.
Qed.

Lemma dif_vect0:
forall v,
dif_v v vect0 = v. 
Proof.
by move => v; rewrite - ext_eq; rewrite  /dif_v /vect0 =>i;
rewrite !vxK  //=; ring_simplify.
Qed.

Lemma dist_min_tri:
forall a b c,
dist_Rp a b -dist_Rp c b <= dist_Rp (dif_v a b) (dif_v c b).
Proof.
move=> a b c.
rewrite /dist_Rp.
have H: norm  (a -^ b) <= norm ((a -^ b) -^ (c -^ b)) + norm (c -^ b).
pattern (a -^ b) at 1 ; replace (a -^ b) with (((a -^ b) -^ (c -^ b)) +^ (c -^ b)).
apply norm_tri.
by rewrite -ext_eq => i; rewrite !vxK; ring_simplify.
fourier.
Qed.

End vectors.
 
Implicit Arguments base_v [p].
Implicit Arguments norm [p].
Implicit Arguments norm_pos [p].
Implicit Arguments norm_hom_pos [p].
Implicit Arguments norm0 [p].
Implicit Arguments dist_Rp [p].
Implicit Arguments dist_pos [p].
Implicit Arguments dist_refl [p].
Implicit Arguments dist_sym [p].
Implicit Arguments dist_tri [p].
Implicit Arguments dif_v [p].
Implicit Arguments add_v [p].
Implicit Arguments opp_v [p].
Implicit Arguments  mult_sv [p].
Implicit Arguments vec_of_fun [p].
Implicit Arguments norm_rew [p].
Implicit Arguments norm_tri [p].


(*
Implicit Arguments dist_Rp [p].
Implicit Arguments dist_Rp [p].
*)